"""
I/O for the TetGen file format, c.f.
<https://wias-berlin.de/software/tetgen/fformats.node.html>
"""
import pathlib

import numpy as np

from ..__about__ import __version__
from .._common import warn
from .._exceptions import ReadError, WriteError
from .._helpers import register_format
from .._mesh import CellBlock, Mesh


def read(filename):
    filename = pathlib.Path(filename)
    if filename.suffix == ".node":
        node_filename = filename
        ele_filename = filename.parent / (filename.stem + ".ele")
    elif filename.suffix == ".ele":
        node_filename = filename.parent / (filename.stem + ".node")
        ele_filename = filename
    else:
        raise ReadError()

    point_data = {}
    cell_data = {}

    # read nodes
    with open(node_filename) as f:
        line = f.readline().strip()
        while len(line) == 0 or line[0] == "#":
            line = f.readline().strip()

        num_points, dim, num_attrs, num_bmarkers = (
            int(item) for item in line.split(" ") if item != ""
        )
        if dim != 3:
            raise ReadError("Need 3D points.")

        points = np.fromfile(
            f, dtype=float, count=(4 + num_attrs + num_bmarkers) * num_points, sep=" "
        ).reshape(num_points, 4 + num_attrs + num_bmarkers)

        node_index_base = int(points[0, 0])
        # make sure the nodes are numbered consecutively
        if not np.all(
            points[:, 0]
            == np.arange(node_index_base, node_index_base + points.shape[0])
        ):
            raise ReadError()
        # read point attributes
        for k in range(num_attrs):
            point_data[f"tetgen:attr{k + 1}"] = points[:, 4 + k]
        # read boundary markers, the first is "ref", the others are "ref2", "ref3", ...
        for k in range(num_bmarkers):
            flag = "" if k == 0 else str(k + 1)
            point_data["tetgen:ref" + flag] = points[:, 4 + num_attrs + k]
        # remove the leading index column, the attributes, and the boundary markers
        points = points[:, 1:4]

    # read elements
    with open(ele_filename.as_posix()) as f:
        line = f.readline().strip()
        while len(line) == 0 or line[0] == "#":
            line = f.readline().strip()

        num_tets, num_points_per_tet, num_attrs = (
            int(item) for item in line.strip().split(" ") if item != ""
        )
        if num_points_per_tet != 4:
            raise ReadError()
        cells = np.fromfile(
            f, dtype=int, count=(5 + num_attrs) * num_tets, sep=" "
        ).reshape(num_tets, 5 + num_attrs)
        # read cell (region) attributes, the first is "ref", the others are "ref2",
        # "ref3", ...
        for k in range(num_attrs):
            flag = "" if k == 0 else str(k + 1)
            cell_data["tetgen:ref" + flag] = [cells[:, 5 + k]]
        # remove the leading index column and the attributes
        cells = cells[:, 1:5]
        cells -= node_index_base

    return Mesh(
        points, [CellBlock("tetra", cells)], point_data=point_data, cell_data=cell_data
    )


def write(filename, mesh, float_fmt=".16e"):
    filename = pathlib.Path(filename)
    if filename.suffix == ".node":
        node_filename = filename
        ele_filename = filename.parent / (filename.stem + ".ele")
    elif filename.suffix == ".ele":
        node_filename = filename.parent / (filename.stem + ".node")
        ele_filename = filename
    else:
        raise WriteError(f"Must specify .node or .ele file. Got {filename}.")

    if mesh.points.shape[1] != 3:
        raise WriteError("Can only write 3D points")

    # write nodes
    with open(node_filename, "w") as fh:
        # identify ":ref" key
        attr_keys = list(mesh.point_data.keys())
        ref_keys = [k for k in attr_keys if ":ref" in k]
        if len(attr_keys) > 0:
            if len(ref_keys) > 0:
                ref_keys = ref_keys[:1]
                attr_keys.remove(ref_keys[0])
            else:
                ref_keys = attr_keys[:1]
                attr_keys = attr_keys[1:]

        nattr, nref = len(attr_keys), len(ref_keys)
        fh.write(f"# This file was created by meshio v{__version__}\n")
        if (nattr + nref) > 0:
            fh.write(
                "# attribute and marker names: {}\n".format(
                    ", ".join(attr_keys + ref_keys)
                )
            )
        fh.write(f"{mesh.points.shape[0]} {3} {nattr} {nref}\n")
        fmt = (
            "{} "
            + " ".join((3 + nattr) * ["{:" + float_fmt + "}"])
            + "".join((nref) * [" {}"])
            + "\n"
        )
        for k, pt in enumerate(mesh.points):
            data = (
                list(pt[:3])
                + [mesh.point_data[key][k] for key in attr_keys]
                + [mesh.point_data[key][k] for key in ref_keys]
            )
            fh.write(fmt.format(k, *data))

    if any(c.type != "tetra" for c in mesh.cells):
        string = ", ".join([c.type for c in mesh.cells if c.type != "tetra"])
        warn(f"TetGen only supports tetrahedra, but mesh has {string}. Skipping those.")

    # write cells
    with open(ele_filename, "w") as fh:
        attr_keys = list(mesh.cell_data.keys())
        ref_keys = [k for k in attr_keys if ":ref" in k]
        if len(attr_keys) > 0:
            if len(ref_keys) > 0:
                attr_keys.remove(ref_keys[0])
                attr_keys = ref_keys[:1] + attr_keys

        nattr = len(attr_keys)
        fh.write(f"# This file was created by meshio v{__version__}\n")
        if nattr > 0:
            fh.write("# attribute names: {}\n".format(", ".join(attr_keys)))
        for id, c in enumerate(filter(lambda c: c.type == "tetra", mesh.cells)):
            data = c.data
            fh.write(f"{data.shape[0]} {4} {nattr}\n")
            fmt = " ".join((5 + nattr) * ["{}"]) + "\n"
            for k, tet in enumerate(data):
                data = list(tet[:4]) + [mesh.cell_data[key][id][k] for key in attr_keys]
                fh.write(fmt.format(k, *data))


register_format("tetgen", [".ele", ".node"], read, {"tetgen": write})



"""
I/O for PERMAS dat files.
"""
import numpy as np

from ..__about__ import __version__
from .._common import warn
from .._exceptions import ReadError
from .._files import open_file
from .._helpers import register_format
from .._mesh import CellBlock, Mesh

permas_to_meshio_type = {
    "PLOT1": "vertex",
    "PLOTL2": "line",
    "FLA2": "line",
    "FLA3": "line3",
    "PLOTL3": "line3",
    "BECOS": "line",
    "BECOC": "line",
    "BETAC": "line",
    "BECOP": "line",
    "BETOP": "line",
    "BEAM2": "line",
    "FSCPIPE2": "line",
    "LOADA4": "quad",
    "PLOTA4": "quad",
    "QUAD4": "quad",
    "QUAD4S": "quad",
    "QUAMS4": "quad",
    "SHELL4": "quad",
    "PLOTA8": "quad8",
    "LOADA8": "quad8",
    "QUAMS8": "quad8",
    "PLOTA9": "quad9",
    "LOADA9": "quad9",
    "QUAMS9": "quad9",
    "PLOTA3": "triangle",
    "SHELL3": "triangle",
    "TRIA3": "triangle",
    "TRIA3K": "triangle",
    "TRIA3S": "triangle",
    "TRIMS3": "triangle",
    "LOADA6": "triangle6",
    "TRIMS6": "triangle6",
    "HEXE8": "hexahedron",
    "HEXFO8": "hexahedron",
    "HEXE20": "hexahedron20",
    "HEXE27": "hexahedron27",
    "TET4": "tetra",
    "TET10": "tetra10",
    "PYRA5": "pyramid",
    "PENTA6": "wedge",
    "PENTA15": "wedge15",
}
meshio_to_permas_type = {v: k for k, v in permas_to_meshio_type.items()}


def read(filename):
    """Reads a PERMAS dat file."""
    with open_file(filename, "r") as f:
        out = read_buffer(f)
    return out


def read_buffer(f):
    # Initialize the optional data fields
    cells = []
    nsets = {}
    elsets = {}
    field_data = {}
    cell_data = {}
    point_data = {}

    while True:
        line = f.readline()
        if not line:
            # EOF
            break

        # Comments
        if line.startswith("!"):
            continue

        keyword = line.strip("$").upper()
        if keyword.startswith("COOR"):
            points, point_gids = _read_nodes(f)
        elif keyword.startswith("ELEMENT"):
            key, idx = _read_cells(f, keyword, point_gids)
            cells.append(CellBlock(key, idx))
        elif keyword.startswith("NSET"):
            params_map = get_param_map(keyword, required_keys=["NSET"])
            setids = read_set(f, params_map)
            name = params_map["NSET"]
            if name not in nsets:
                nsets[name] = []
            nsets[name].append(setids)
        elif keyword.startswith("ESET"):
            params_map = get_param_map(keyword, required_keys=["ESET"])
            setids = read_set(f, params_map)
            name = params_map["ESET"]
            if name not in elsets:
                elsets[name] = []
            elsets[name].append(setids)
        else:
            # There are just too many PERMAS keywords to explicitly skip them.
            pass

    return Mesh(
        points, cells, point_data=point_data, cell_data=cell_data, field_data=field_data
    )


def _read_nodes(f):
    points = []
    point_gids = {}
    index = 0
    while True:
        last_pos = f.tell()
        line = f.readline()
        if line.startswith("!"):
            break
        if line.startswith("$"):
            break
        entries = line.strip().split(" ")
        gid, x = entries[0], entries[1:]
        point_gids[int(gid)] = index
        points.append([float(xx) for xx in x])
        index += 1

    f.seek(last_pos)
    return np.array(points, dtype=float), point_gids


def _read_cells(f, line0, point_gids):
    sline = line0.split(" ")[1:]
    etype_sline = sline[0]
    if "TYPE" not in etype_sline:
        raise ReadError(etype_sline)
    etype = etype_sline.split("=")[1].strip()
    if etype not in permas_to_meshio_type:
        raise ReadError(f"Element type not available: {etype}")
    cell_type = permas_to_meshio_type[etype]
    cells, idx = [], []
    while True:
        last_pos = f.tell()
        line = f.readline()
        if line.startswith("$") or line == "":
            break
        line = line.strip()
        # the first item is just a running index
        idx += [point_gids[int(k)] for k in filter(None, line.split(" ")[1:])]
        if not line.endswith("!"):
            cells.append(idx)
            idx = []
    f.seek(last_pos)
    return cell_type, np.array(cells)


def get_param_map(word, required_keys=None):
    """
    get the optional arguments on a line

    Example
    -------
    >>> iline = 0
    >>> word = 'elset,instance=dummy2,generate'
    >>> params = get_param_map(iline, word, required_keys=['instance'])
    params = {
        'elset' : None,
        'instance' : 'dummy2,
        'generate' : None,
    }
    """
    if required_keys is None:
        required_keys = []
    words = word.split(",")
    param_map = {}
    for wordi in words:
        if "=" not in wordi:
            key = wordi.strip()
            value = None
        else:
            sword = wordi.split("=")
            if len(sword) != 2:
                raise ReadError(sword)
            key = sword[0].strip()
            value = sword[1].strip()
        param_map[key] = value

    msg = ""
    for key in required_keys:
        if key not in param_map:
            msg += f"{key} not found in {word}\n"
    if msg:
        raise RuntimeError(msg)
    return param_map


def read_set(f, params_map):
    set_ids = []
    while True:
        last_pos = f.tell()
        line = f.readline()
        if line.startswith("$"):
            break
        set_ids += [int(k) for k in line.strip().strip(" ").split(" ")]
    f.seek(last_pos)

    if "generate" in params_map:
        if len(set_ids) != 3:
            raise ReadError(set_ids)
        set_ids = np.arange(set_ids[0], set_ids[1], set_ids[2])
    else:
        try:
            set_ids = np.unique(np.array(set_ids, dtype="int32"))
        except ValueError:
            raise
    return set_ids


def write(filename, mesh):
    if mesh.points.shape[1] == 2:
        warn(
            "PERMAS requires 3D points, but 2D points given. "
            "Appending 0 third component."
        )
        points = np.column_stack([mesh.points, np.zeros_like(mesh.points[:, 0])])
    else:
        points = mesh.points

    with open_file(filename, "wt") as f:
        f.write("!PERMAS DataFile Version 18.0\n")
        f.write(f"!written by meshio v{__version__}\n")
        f.write("$ENTER COMPONENT NAME=DFLT_COMP\n")
        f.write("$STRUCTURE\n")
        f.write("$COOR\n")
        for k, x in enumerate(points):
            f.write(f"{k + 1} {x[0]} {x[1]} {x[2]}\n")
        eid = 0
        tria6_order = [0, 3, 1, 4, 2, 5]
        tet10_order = [0, 4, 1, 5, 2, 6, 7, 8, 9, 3]
        quad9_order = [0, 4, 1, 7, 8, 5, 3, 6, 2]
        wedge15_order = [0, 6, 1, 7, 2, 8, 9, 10, 11, 3, 12, 4, 13, 5, 14]
        for cell_block in mesh.cells:
            node_idcs = cell_block.data
            f.write("!\n")
            f.write("$ELEMENT TYPE=" + meshio_to_permas_type[cell_block.type] + "\n")
            if cell_block.type == "tetra10":
                for row in node_idcs:
                    eid += 1
                    mylist = row.tolist()
                    mylist = [mylist[i] for i in tet10_order]
                    nids_strs = (str(nid + 1) for nid in mylist)
                    f.write(str(eid) + " " + " ".join(nids_strs) + "\n")
            elif cell_block.type == "triangle6":
                for row in node_idcs:
                    eid += 1
                    mylist = row.tolist()
                    mylist = [mylist[i] for i in tria6_order]
                    nids_strs = (str(nid + 1) for nid in mylist)
                    f.write(str(eid) + " " + " ".join(nids_strs) + "\n")
            elif cell_block.type == "quad9":
                for row in node_idcs:
                    eid += 1
                    mylist = row.tolist()
                    mylist = [mylist[i] for i in quad9_order]
                    nids_strs = (str(nid + 1) for nid in mylist)
                    f.write(str(eid) + " " + " ".join(nids_strs) + "\n")
            elif cell_block.type == "wedge15":
                for row in node_idcs:
                    eid += 1
                    mylist = row.tolist()
                    mylist = [mylist[i] for i in wedge15_order]
                    nids_strs = (str(nid + 1) for nid in mylist)
                    f.write(str(eid) + " " + " ".join(nids_strs) + "\n")
            else:
                for row in node_idcs:
                    eid += 1
                    nids_strs = (str(nid + 1) for nid in row.tolist())
                    f.write(str(eid) + " " + " ".join(nids_strs) + "\n")
        f.write("$END STRUCTURE\n")
        f.write("$EXIT COMPONENT\n")
        f.write("$FIN\n")


register_format(
    "permas", [".post", ".post.gz", ".dato", ".dato.gz"], read, {"permas": write}
)



import numpy as np
from pathlib import Path
import json

from foilmesh.splinerefine import SplineRefine
from foilmesh.trailingedge import TrailingEdge
from foilmesh import meshing as meshing

from foilmesh import connect
from foilmesh.controldict import ControlDict, extension

import logging
from typing import Union

from foilmesh.utils import timer

logger = logging.getLogger(__name__)


class Airfoil:
    """Class to read airfoil data from file (or use predefined airfoil)

    Attributes:
        penwidth (float): thickness of airfoil outline
        raw_coordinates (numpy array): list of contour points as tuples
    """

    def __init__(self, name):
        # get MainWindow instance (overcomes handling parents)
        # self.mainwindow = QtCore.QCoreApplication.instance().mainwindow

        self.name = name
        self.chord = None
        self.has_TE = False
        # self.contourPolygon = None
        # self.contourSpline = None
        self.spline_data = None
        self.raw_coordinates = None
        # self.penwidth = 4.0
        self.control = ControlDict

    def readContour(self, filename: Union[str, np.ndarray]):
        if isinstance(filename, np.ndarray):
            filename = filename.astype("str").tolist()
            data = [" ".join(i) for i in filename]
        else:
            try:
                with open(filename, mode="r") as f:
                    lines = f.readlines()
            except IOError as error:
                # exc_info=True sends traceback to the logger
                logger.error(
                    "Failed to open file {} with error {}".format(filename, error),
                    exc_info=True,
                )
                return False

            data = [line.strip() for line in lines[1:] if line.strip() != ""]

        # find and drop duplicate points (except first and last)
        data_clean = list()
        for index, line in enumerate(data):
            if index == 0:
                data_clean.append(line)
                continue
            elif index == len(data) - 1:
                data_clean.append(line)
                break

            if line != data[index - 1]:
                data_clean.append(line)
            else:
                logger.info("Dropped duplicate point {}".format(line))

        # check for correct data
        # specifically important for drag and drop
        try:
            x = [float(l.split()[0]) for l in data_clean]
            y = [float(l.split()[1]) for l in data_clean]
        except (ValueError, IndexError) as error:
            logger.error("Unable to parse file file {}".format(filename))
            logger.error("Following error occured: {}".format(error))
            return False
        except:
            # exc_info=True sends traceback to the logger
            logger.error(
                "Unable to parse file file {}. Unknown error caught".format(filename),
                exc_info=True,
            )
            return False

        # store airfoil coordinates as list of tuples
        self.raw_coordinates = np.array((x, y))

        # normalize airfoil to unit chord
        self.raw_coordinates[0] -= np.min(x)
        divisor = np.max(self.raw_coordinates[0])
        self.raw_coordinates[0] /= divisor
        self.raw_coordinates[1] /= divisor

        self.offset = [np.min(y), np.max(y)]
        return True

    def readControl(self, jsonfile):
        with open(jsonfile, "r") as fin:
            data = json.load(fin)
        self.control.Airfoil_file = data["Airfoil_file"]
        self.control.Output = data["Output"]
        self.control.Spline_refine.update(data["Spline_refine"])
        self.control.Trailing_edges = data["Trailing_edges"]
        self.control.Trailing_control.update(data["Trailing_control"])
        self.control.Airfoil_mesh.update(data["Airfoil_mesh"])
        self.control.Trailing_mesh.update(data["Trailing_mesh"])
        self.control.Windtunnel_airfoil.update(data["Windtunnel_airfoil"])
        self.control.Windtunnel_wake.update(data["Windtunnel_wake"])

    @timer
    def StructureMesh(self):
        if (
            self.control.Airfoil_data is not None
            and self.control.Airfoil_file is not None
        ):
            logging.error("Airdoil_file and Airfoil_data only one can be defined!")
        else:
            print(f'---> Airfoil "{self.name}" to mesh --->')
            print("---> Start mesh structure gird --->")
            if self.control.Airfoil_file is not None:
                self.readContour(self.control.Airfoil_file)
            else:
                self.readContour(self.control.Airfoil_data)
            refine = SplineRefine(self)
            refine.doSplineRefine(
                tolerance=self.control.Spline_refine["Refinement tolerance"],
                points=self.control.Spline_refine["Number of points on spline"],
                ref_te=self.control.Spline_refine["Refine trailing edge old"],
                ref_te_n=self.control.Spline_refine["Refine trailing edge new"],
                ref_te_ratio=self.control.Spline_refine["Refine trailing edge ratio"],
            )

            if self.control.Trailing_edges:
                self.has_TE = True
                te = self.control.Trailing_control
                trailing = TrailingEdge(self)
                trailing.trailingEdge(
                    blend=te["Upper side blending length"] / 100.0,
                    ex=te["Upper blending polynomial exponent"],
                    thickness=te["Trailing edge thickness relative to chord"],
                    side="upper",
                )
                trailing.trailingEdge(
                    blend=te["Lower side blending length"] / 100.0,
                    ex=te["Lower blending polynomial exponent"],
                    thickness=te["Trailing edge thickness relative to chord"],
                    side="lower",
                )

            wind_tunnel = meshing.Windtunnel(self)
            contour = self.spline_data[0]

            # mesh around airfoil
            acm = self.control.Airfoil_mesh
            wind_tunnel.AirfoilMesh(
                name="block_airfoil",
                contour=contour,
                divisions=acm["Divisions normal to airfoil"],
                ratio=acm["Cell growth rate"],
                thickness=acm["1st cell layer thickness"],
            )

            # mesh at trailing edge
            tem = self.control.Trailing_mesh
            wind_tunnel.TrailingEdgeMesh(
                name="block_TE",
                te_divisions=tem["Divisions at trailing edge"],
                thickness=tem["1st cell layer thickness"],
                divisions=tem["Divisions downstream"],
                ratio=tem["Cell growth rate"],
            )

            # mesh tunnel airfoil
            tam = self.control.Windtunnel_airfoil
            wind_tunnel.TunnelMesh(
                name="block_tunnel",
                tunnel_height=tam["Windtunnel height"],
                divisions_height=tam["Divisions of tunnel height"],
                ratio_height=tam["Cell thickness ratio"],
                dist=tam["Distribution biasing"],
                smoothing_algorithm=tam["Smoothing algorithm"],
                smoothing_iterations=tam["Smoothing iterations"],
                smoothing_tolerance=tam["Smoothing tolerance"],
            )

            # mesh tunnel wake
            twm = self.control.Windtunnel_wake
            wind_tunnel.TunnelMeshWake(
                name="block_tunnel_wake",
                tunnel_wake=twm["Windtunnel wake"],
                divisions=twm["Divisions in the wake"],
                ratio=twm["Cell thickness ratio"],
                spread=twm["Equalize vertical wake line at"] / 100.0,
            )
            cnt=connect.Connect(None)
            # connect = connect.Connect(None)
            vertices, connectivity, _ = cnt.connectAllBlocks(wind_tunnel.blocks)

            # add mesh to Wind-tunnel instance
            wind_tunnel.mesh = vertices, connectivity

            # generate cell to edge connectivity from mesh
            wind_tunnel.makeLCE()

            # generate cell to edge connectivity from mesh
            wind_tunnel.makeLCE()

            # generate boundaries from mesh connectivity
            wind_tunnel.makeBoundaries()

            print("<--- Finished meshing <---")
            print("---> Starting mesh export --->")

            mesh_name = Path(self.control.Output)
            mesh_folder=mesh_name.parent
            if not mesh_folder.exists():
                mesh_folder.mkdir(parents=True,exist_ok=True)
            getattr(meshing.BlockMesh, "write" + extension[mesh_name.suffix])(
                wind_tunnel, name=mesh_name
            )

            print(f"<--- Finished \"{mesh_name}\" export <---")


if __name__ == "__main__":
    import foilmesh as fm

    af = fm.Airfoil(name="myairfoil")
    # af.control.Airfoil_data=np.array([[1.00,0.95],[0.89565,0.777]])
    af.readControl("./example.json")
    af.StructureMesh()



import numpy as np

from .._helpers import _writer_map, read, reader_map, write


def add_args(parser):
    parser.add_argument("infile", type=str, help="mesh file to be read from")
    parser.add_argument(
        "--input-format",
        "-i",
        type=str,
        choices=sorted(list(reader_map.keys())),
        help="input file format",
        default=None,
    )
    parser.add_argument(
        "--output-format",
        "-o",
        type=str,
        choices=sorted(list(_writer_map.keys())),
        help="output file format",
        default=None,
    )
    parser.add_argument(
        "--ascii",
        "-a",
        action="store_true",
        help="write in ASCII format variant (where applicable, default: binary)",
    )
    parser.add_argument("outfile", type=str, help="mesh file to be written to")
    parser.add_argument(
        "--float-format",
        "-f",
        type=str,
        help="float format used in output ASCII files (default: .16e)",
    )
    parser.add_argument(
        "--sets-to-int-data",
        "-s",
        action="store_true",
        help="if possible, convert sets to integer data (useful if the output type does not support sets)",
    )
    parser.add_argument(
        "--int-data-to-sets",
        "-d",
        action="store_true",
        help="if possible, convert integer data to sets (useful if the output type does not support integer data)",
    )


def convert(args):
    # read mesh data
    mesh = read(args.infile, file_format=args.input_format)

    # Some converters (like VTK) require `points` to be contiguous.
    mesh.points = np.ascontiguousarray(mesh.points)

    if args.sets_to_int_data:
        mesh.point_sets_to_data()
        mesh.cell_sets_to_data()

    if args.int_data_to_sets:
        for key in mesh.point_data:
            mesh.point_data_to_sets(key)
        for key in mesh.cell_data:
            mesh.cell_data_to_sets(key)

    # write it out
    kwargs = {"file_format": args.output_format}
    if args.float_format is not None:
        kwargs["float_fmt"] = args.float_format
    if args.ascii:
        kwargs["binary"] = False

    write(args.outfile, mesh, **kwargs)



"""
I/O for Gmsh's msh format (version 4.1, as used by Gmsh 4.2.2+), cf.
<http://gmsh.info/doc/texinfo/gmsh.html#MSH-file-format>.
"""
from functools import partial

import numpy as np

from .._common import cell_data_from_raw, num_nodes_per_cell, raw_from_cell_data, warn
from .._exceptions import ReadError, WriteError
from .._mesh import CellBlock, Mesh
from .common import (
    _fast_forward_over_blank_lines,
    _fast_forward_to_end_block,
    _gmsh_to_meshio_order,
    _gmsh_to_meshio_type,
    _meshio_to_gmsh_order,
    _meshio_to_gmsh_type,
    _read_data,
    _read_physical_names,
    _write_data,
    _write_physical_names,
)

c_int = np.dtype("i")
c_size_t = np.dtype("P")
c_double = np.dtype("d")


def _size_type(data_size):
    return np.dtype(f"u{data_size}")


def read_buffer(f, is_ascii: bool, data_size):
    # The format is specified at
    # <http://gmsh.info/doc/texinfo/gmsh.html#MSH-file-format>.

    # Initialize the optional data fields
    points = []
    cells = None
    field_data = {}
    cell_data_raw = {}
    cell_tags = {}
    point_data = {}
    physical_tags = None
    bounding_entities = None
    cell_sets = {}
    periodic = None
    while True:
        # fast-forward over blank lines
        line, is_eof = _fast_forward_over_blank_lines(f)
        if is_eof:
            break

        if line[0] != "$":
            raise ReadError(f"Unexpected line {repr(line)}")

        environ = line[1:].strip()

        if environ == "PhysicalNames":
            _read_physical_names(f, field_data)
        elif environ == "Entities":
            # Read physical tags and information on bounding entities.
            # The information is passed to the processing of elements.
            physical_tags, bounding_entities = _read_entities(f, is_ascii, data_size)
        elif environ == "Nodes":
            points, point_tags, point_entities = _read_nodes(f, is_ascii, data_size)
        elif environ == "Elements":
            cells, cell_tags, cell_sets = _read_elements(
                f,
                point_tags,
                physical_tags,
                bounding_entities,
                is_ascii,
                data_size,
                field_data,
            )
        elif environ == "Periodic":
            periodic = _read_periodic(f, is_ascii, data_size)
        elif environ == "NodeData":
            _read_data(f, "NodeData", point_data, data_size, is_ascii)
        elif environ == "ElementData":
            _read_data(f, "ElementData", cell_data_raw, data_size, is_ascii)
        else:
            # From
            # <http://gmsh.info/doc/texinfo/gmsh.html#MSH-file-format>:
            # ```
            # Any section with an unrecognized header is simply ignored: you can thus
            # add comments in a .msh file by putting them e.g. inside a
            # $Comments/$EndComments section.
            # ```
            # skip environment
            _fast_forward_to_end_block(f, environ)

    if cells is None:
        raise ReadError("$Element section not found.")
    cell_data = cell_data_from_raw(cells, cell_data_raw)
    cell_data.update(cell_tags)

    # Add node entity information to the point data
    point_data.update({"gmsh:dim_tags": point_entities})

    return Mesh(
        points,
        cells,
        point_data=point_data,
        cell_data=cell_data,
        field_data=field_data,
        cell_sets=cell_sets,
        gmsh_periodic=periodic,
    )


def _read_entities(f, is_ascii: bool, data_size):
    # Read the entity section. Return physical tags of the entities, and (for
    # entities of dimension > 0) the bounding entities (so points that form
    # the boundary of a line etc).
    # Note that the bounding box of the entities is disregarded. Adding this
    # is not difficult, but for the moment, the entropy of adding more data
    # does not seem warranted.

    fromfile = partial(np.fromfile, sep=" " if is_ascii else "")
    c_size_t = _size_type(data_size)
    physical_tags = ({}, {}, {}, {})
    bounding_entities = ({}, {}, {}, {})
    number = fromfile(f, c_size_t, 4)  # dims 0, 1, 2, 3
    for d, n in enumerate(number):
        for _ in range(n):
            (tag,) = fromfile(f, c_int, 1)
            fromfile(f, c_double, 3 if d == 0 else 6)  # discard bounding-box
            (num_physicals,) = fromfile(f, c_size_t, 1)
            physical_tags[d][tag] = list(fromfile(f, c_int, num_physicals))
            if d > 0:
                # Number of bounding entities
                num_BREP_ = fromfile(f, c_size_t, 1)[0]
                # Store bounding entities
                bounding_entities[d][tag] = fromfile(f, c_int, num_BREP_)

    _fast_forward_to_end_block(f, "Entities")
    return physical_tags, bounding_entities


def _read_nodes(f, is_ascii: bool, data_size):
    # Read node data: Node coordinates and tags.
    # Also find the entities of the nodes, and store this as point_data.
    # Note that entity tags are 1-offset within each dimension, thus it is
    # necessary to keep track of both tag and dimension of the entity

    fromfile = partial(np.fromfile, sep=" " if is_ascii else "")
    c_size_t = _size_type(data_size)

    # numEntityBlocks numNodes minNodeTag maxNodeTag (all size_t)
    num_entity_blocks, total_num_nodes, _, _ = fromfile(f, c_size_t, 4)

    points = np.empty((total_num_nodes, 3), dtype=float)
    tags = np.empty(total_num_nodes, dtype=int)
    dim_tags = np.empty((total_num_nodes, 2), dtype=int)

    # To save the entity block id for each node, initialize an array here,
    # populate it with num_nodes
    idx = 0
    for _ in range(num_entity_blocks):
        # entityDim(int) entityTag(int) parametric(int) numNodes(size_t)
        dim, entity_tag, parametric = fromfile(f, c_int, 3)
        if parametric != 0:
            raise ReadError("parametric nodes not implemented")
        num_nodes = int(fromfile(f, c_size_t, 1)[0])

        # From <http://gmsh.info/doc/texinfo/gmsh.html#MSH-file-format>:
        # > [...] tags can be "sparse", i.e., do not have to constitute a continuous
        # > list of numbers (the format even allows them to not be ordered).
        #
        # Following https://github.com/nschloe/meshio/issues/388, we read the tags and
        # populate the points array accordingly, thereby preserving the order of indices
        # of nodes/points.
        ixx = slice(idx, idx + num_nodes)
        tags[ixx] = fromfile(f, c_size_t, num_nodes) - 1

        # Store the point densely and in the order in which they appear in the file.
        # x(double) y(double) z(double) (* numNodes)
        points[ixx] = fromfile(f, c_double, num_nodes * 3).reshape((num_nodes, 3))

        # Entity tag and entity dimension of the nodes. Stored as point-data.
        dim_tags[ixx, 0] = dim
        dim_tags[ixx, 1] = entity_tag
        idx += num_nodes

    _fast_forward_to_end_block(f, "Nodes")

    return points, tags, dim_tags


def _read_elements(
    f, point_tags, physical_tags, bounding_entities, is_ascii, data_size, field_data
):
    fromfile = partial(np.fromfile, sep=" " if is_ascii else "")
    c_size_t = _size_type(data_size)

    # numEntityBlocks numElements minElementTag maxElementTag (all size_t)
    num_entity_blocks, _, _, _ = fromfile(f, c_size_t, 4)

    data = []
    cell_data = {}
    cell_sets = {k: [None] * num_entity_blocks for k in field_data.keys()}

    for k in range(num_entity_blocks):
        # entityDim(int) entityTag(int) elementType(int) numElements(size_t)
        dim, tag, type_ele = fromfile(f, c_int, 3)
        (num_ele,) = fromfile(f, c_size_t, 1)
        for physical_name, cell_set in cell_sets.items():
            cell_set[k] = np.arange(
                num_ele
                if (
                    physical_tags
                    and field_data[physical_name][1] == dim
                    and field_data[physical_name][0] in physical_tags[dim][tag]
                )
                else 0,
                dtype=type(num_ele),
            )
        tpe = _gmsh_to_meshio_type[type_ele]
        num_nodes_per_ele = num_nodes_per_cell[tpe]
        d = fromfile(f, c_size_t, int(num_ele * (1 + num_nodes_per_ele))).reshape(
            (num_ele, -1)
        )

        # Find physical tag, if defined; else it is None.
        pt = None if not physical_tags else physical_tags[dim][tag]
        # Bounding entities (of lower dimension) if defined. Else it is None.
        if dim > 0 and bounding_entities:  # Points have no boundaries
            be = bounding_entities[dim][tag]
        else:
            be = None
        data.append((pt, be, tag, tpe, d))

    _fast_forward_to_end_block(f, "Elements")

    # Inverse point tags
    inv_tags = np.full(np.max(point_tags) + 1, -1, dtype=int)
    inv_tags[point_tags] = np.arange(len(point_tags))

    # Note that the first column in the data array is the element tag; discard it.
    data = [
        (physical_tag, bound_entity, geom_tag, tpe, inv_tags[d[:, 1:] - 1])
        for physical_tag, bound_entity, geom_tag, tpe, d in data
    ]

    cells = []
    for physical_tag, bound_entity, geom_tag, key, values in data:
        cells.append(CellBlock(key, _gmsh_to_meshio_order(key, values)))
        if physical_tag:
            if "gmsh:physical" not in cell_data:
                cell_data["gmsh:physical"] = []
            cell_data["gmsh:physical"].append(
                np.full(len(values), physical_tag[0], int)
            )
        if "gmsh:geometrical" not in cell_data:
            cell_data["gmsh:geometrical"] = []
        cell_data["gmsh:geometrical"].append(np.full(len(values), geom_tag, int))

        # The bounding entities is stored in the cell_sets.
        if bounding_entities:
            if "gmsh:bounding_entities" not in cell_sets:
                cell_sets["gmsh:bounding_entities"] = []
            cell_sets["gmsh:bounding_entities"].append(bound_entity)

    return cells, cell_data, cell_sets


def _read_periodic(f, is_ascii, data_size):
    fromfile = partial(np.fromfile, sep=" " if is_ascii else "")
    c_size_t = _size_type(data_size)
    periodic = []
    # numPeriodicLinks(size_t)
    num_periodic = int(fromfile(f, c_size_t, 1)[0])
    for _ in range(num_periodic):
        # entityDim(int) entityTag(int) entityTagMaster(int)
        edim, stag, mtag = fromfile(f, c_int, 3)
        # numAffine(size_t) value(double) ...
        num_affine = int(fromfile(f, c_size_t, 1)[0])
        affine = fromfile(f, c_double, num_affine)
        # numCorrespondingNodes(size_t)
        num_nodes = int(fromfile(f, c_size_t, 1)[0])
        # nodeTag(size_t) nodeTagMaster(size_t) ...
        slave_master = fromfile(f, c_size_t, num_nodes * 2).reshape(-1, 2)
        slave_master = slave_master - 1  # Subtract one, Python is 0-based
        periodic.append([edim, (stag, mtag), affine, slave_master])

    _fast_forward_to_end_block(f, "Periodic")
    return periodic


def write(filename, mesh, float_fmt=".16e", binary=True):
    """Writes msh files, cf.
    <http://gmsh.info/doc/texinfo/gmsh.html#MSH-file-format>.
    """
    # Filter the point data: gmsh:dim_tags are tags, the rest is actual point data.
    point_data = {}
    for key, d in mesh.point_data.items():
        if key not in ["gmsh:dim_tags"]:
            point_data[key] = d

    # Split the cell data: gmsh:physical and gmsh:geometrical are tags, the rest is
    # actual cell data.
    tag_data = {}
    cell_data = {}
    for key, d in mesh.cell_data.items():
        if key in ["gmsh:physical", "gmsh:geometrical", "cell_tags"]:
            tag_data[key] = d
        else:
            cell_data[key] = d

    with open(filename, "wb") as fh:
        file_type = 1 if binary else 0
        data_size = c_size_t.itemsize
        fh.write(b"$MeshFormat\n")
        fh.write(f"4.1 {file_type} {data_size}\n".encode())
        if binary:
            np.array([1], dtype=c_int).tofile(fh)
            fh.write(b"\n")
        fh.write(b"$EndMeshFormat\n")

        if mesh.field_data:
            _write_physical_names(fh, mesh.field_data)

        _write_entities(
            fh, mesh.cells, tag_data, mesh.cell_sets, mesh.point_data, binary
        )
        _write_nodes(fh, mesh.points, mesh.cells, mesh.point_data, float_fmt, binary)
        _write_elements(fh, mesh.cells, tag_data, binary)
        if mesh.gmsh_periodic is not None:
            _write_periodic(fh, mesh.gmsh_periodic, float_fmt, binary)

        for name, dat in point_data.items():
            _write_data(fh, "NodeData", name, dat, binary)
        cell_data_raw = raw_from_cell_data(cell_data)
        for name, dat in cell_data_raw.items():
            _write_data(fh, "ElementData", name, dat, binary)


def _write_entities(fh, cells, tag_data, cell_sets, point_data, binary):
    """Write entity section in a .msh file.

    The entity section links up to three kinds of information:
        1) The geometric objects represented in the mesh.
        2) Physical tags of geometric objects. This data will be a subset
           of that represented in 1)
        3) Which geometric objects form the boundary of this object.
           The boundary is formed of objects with dimension 1 less than
           the current one. A boundary can only be specified for objects of
           dimension at least 1.

    The entities of all geometric objects is pulled from
    point_data['gmsh:dim_tags']. For details, see the function _write_nodes().

    Physical tags are specified as tag_data, while the boundary of a geometric
    object is specified in cell_sets.

    """

    # The data format for the entities section is
    #
    #    numPoints(size_t) numCurves(size_t)
    #      numSurfaces(size_t) numVolumes(size_t)
    #    pointTag(int) X(double) Y(double) Z(double)
    #      numPhysicalTags(size_t) physicalTag(int) ...
    #    ...
    #    curveTag(int) minX(double) minY(double) minZ(double)
    #      maxX(double) maxY(double) maxZ(double)
    #      numPhysicalTags(size_t) physicalTag(int) ...
    #      numBoundingPoints(size_t) pointTag(int) ...
    #    ...
    #    surfaceTag(int) minX(double) minY(double) minZ(double)
    #      maxX(double) maxY(double) maxZ(double)
    #      numPhysicalTags(size_t) physicalTag(int) ...
    #      numBoundingCurves(size_t) curveTag(int) ...
    #    ...
    #    volumeTag(int) minX(double) minY(double) minZ(double)
    #      maxX(double) maxY(double) maxZ(double)
    #      numPhysicalTags(size_t) physicalTag(int) ...
    #      numBoundngSurfaces(size_t) surfaceTag(int) ...

    # Both nodes and cells have entities, but the cell entities are a subset of
    # the nodes. The reason is (if the inner workings of Gmsh has been correctly
    # understood) that node entities are assigned to all
    # objects necessary to specify the geometry whereas only cells of Physical
    # objcets (gmsh jargon) are present among the cell entities.
    # The entities section must therefore be built on the node-entities, if
    # these are available. If this is not the case, we leave this section blank.
    # TODO: Should this give a warning?
    if "gmsh:dim_tags" not in point_data:
        return

    fh.write(b"$Entities\n")

    # Array of entity tag (first row) and dimension (second row) per node.
    # We need to combine the two, since entity tags are reset for each dimension.
    # Uniquify, so that each row in node_dim_tags represent a unique entity
    node_dim_tags = np.unique(point_data["gmsh:dim_tags"], axis=0)

    # Write number of entities per dimension
    num_occ = np.bincount(node_dim_tags[:, 0], minlength=4)
    if num_occ.size > 4:
        raise ValueError("Encountered entity with dimension > 3")

    if binary:
        num_occ.astype(c_size_t).tofile(fh)
    else:
        fh.write(f"{num_occ[0]} {num_occ[1]} {num_occ[2]} {num_occ[3]}\n".encode())

    # Array of dimension and entity tag per cell. Will be compared with the
    # similar not array.
    cell_dim_tags = np.empty((len(cells), 2), dtype=int)
    for ci, cell_block in enumerate(cells):
        cell_dim_tags[ci] = [
            cell_block.dim,
            tag_data["gmsh:geometrical"][ci][0],
        ]

    # We will only deal with bounding entities if this information is available
    has_bounding_elements = "gmsh:bounding_entities" in cell_sets

    # The node entities form a superset of cell entities. Write entity information
    # based on nodes, supplement with cell information when there is a matcihng
    # cell block.
    for dim, tag in node_dim_tags:
        # Find the matching cell block, if it exists
        matching_cell_block = np.where(
            np.logical_and(cell_dim_tags[:, 0] == dim, cell_dim_tags[:, 1] == tag)
        )[0]
        if matching_cell_block.size > 1:
            # It is not 100% clear if this is not permissible, but the current
            # implementation for sure does not allow it.
            raise ValueError("Encountered non-unique CellBlock dim_tag")

        # The information to be written varies according to entity dimension,
        # whether entity has a physical tag, and between ascii and binary.
        # The resulting code is a bit ugly, but no simpler and clean option
        # seems possible.

        # Entity tag
        if binary:
            np.array([tag], dtype=c_int).tofile(fh)
        else:
            fh.write(f"{tag} ".encode())

        # Min-max coordinates for the entity. For now, simply put zeros here,
        # and hope that gmsh does not complain. To expand this, the point
        # coordinates must be made available to this function; the bounding
        # box can then be found by a min-max over the points of the matching
        # cell.
        if dim == 0:
            # Bounding box is a point
            if binary:
                np.zeros(3, dtype=c_double).tofile(fh)
            else:
                fh.write(b"0 0 0 ")
        else:
            # Bounding box has six coordinates
            if binary:
                np.zeros(6, dtype=c_double).tofile(fh)
            else:
                fh.write(b"0 0 0 0 0 0 ")

        # If there is a corresponding cell block, write physical tags (if any)
        # and bounding entities (if any)
        if matching_cell_block.size > 0:
            # entity has a physical tag, write this
            # ASSUMPTION: There is a single physical tag for this
            physical_tag = tag_data["gmsh:physical"][matching_cell_block[0]][0]
            if binary:
                np.array([1], dtype=c_size_t).tofile(fh)
                np.array([physical_tag], dtype=c_int).tofile(fh)
            else:
                fh.write(f"1 {physical_tag} ".encode())
        else:
            # The number of physical tags is zero
            if binary:
                np.array([0], dtype=c_size_t).tofile(fh)
            else:
                fh.write(b"0 ")

        if dim > 0:
            # Entities not of the lowest dimension can have their
            # bounding elements (of dimension one less) specified
            if has_bounding_elements and matching_cell_block.size > 0:
                # The bounding element should be a list
                bounds = cell_sets["gmsh:bounding_entities"][matching_cell_block[0]]
                num_bounds = len(bounds)
                if num_bounds > 0:
                    if binary:
                        np.array(num_bounds, dtype=c_size_t).tofile(fh)
                        np.array(bounds, dtype=c_int).tofile(fh)
                    else:
                        fh.write(f"{num_bounds} ".encode())
                        for bi in bounds:
                            fh.write(f"{bi} ".encode())
                        fh.write(b"\n")
                else:
                    # Register that there are no bounding elements
                    if binary:
                        np.array([0], dtype=c_size_t).tofile(fh)
                    else:
                        fh.write(b"0\n")

            else:
                # Register that there are no bounding elements
                if binary:
                    np.array([0], dtype=c_size_t).tofile(fh)
                else:
                    fh.write(b"0\n")
        else:
            # If ascii, enforce line change
            if not binary:
                fh.write(b"\n")

    if binary:
        fh.write(b"\n")
    # raise NotImplementedError
    fh.write(b"$EndEntities\n")


def _write_nodes(fh, points, cells, point_data, float_fmt, binary):
    """Write node information.

    If data on dimension and tags of the geometric entities which the nodes belong to
    is available available, the nodes will be grouped accordingly. This data is
    specified as point_data, using the key 'gmsh:dim_tags' and data as an
    num_points x 2 numpy array (first column is the dimension of the geometric entity
    of this node, second is the tag).

    If dim_tags are not available, all nodes will be assigned the same tag of 0. This
    only makes sense if a single cell block is present in the mesh; an error will be
    raised if len(cells) > 1.

    """
    if points.shape[1] == 2:
        # msh4 requires 3D points, but 2D points given.
        # Appending 0 third component.
        points = np.column_stack([points, np.zeros_like(points[:, 0])])

    fh.write(b"$Nodes\n")

    # The format for the nodes section is
    #
    # $Nodes
    #   numEntityBlocks(size_t) numNodes(size_t) minNodeTag(size_t) maxNodeTag(size_t)
    #   entityDim(int) entityTag(int) parametric(int; 0 or 1)
    #   numNodesInBlock(size_t)
    #     nodeTag(size_t)
    #     ...
    #     x(double) y(double) z(double)
    #        < u(double; if parametric and entityDim >= 1) >
    #        < v(double; if parametric and entityDim >= 2) >
    #        < w(double; if parametric and entityDim == 3) >
    #     ...
    #   ...
    # $EndNodes
    #
    n = points.shape[0]
    min_tag = 1
    max_tag = n
    is_parametric = 0

    # If node (entity) tag and dimension is available, we make a list of unique
    # combinations thereof, and a map from the full node set to the unique
    # set.
    if "gmsh:dim_tags" in point_data:
        # reverse_index_map maps from all nodes to their respective representation in
        # (the uniquified) node_dim_tags. This approach works for general orderings of
        # the nodes
        node_dim_tags, reverse_index_map = np.unique(
            point_data["gmsh:dim_tags"],
            axis=0,
            return_inverse=True,
        )
    else:
        # If entity information is not provided, we will assign the same entity for all
        # nodes. This only makes sense if the cells are of a single type
        if len(cells) != 1:
            raise WriteError(
                "Specify entity information (gmsh:dim_tags in point_data) "
                + "to deal with more than one cell type. "
            )

        dim = cells[0].dim
        tag = 0
        node_dim_tags = np.array([[dim, tag]])
        # All nodes map to the (single) dimension-entity object
        reverse_index_map = np.full(n, 0, dtype=int)

    num_blocks = node_dim_tags.shape[0]

    # First write preamble
    if binary:
        if points.dtype != c_double:
            warn(f"Binary Gmsh needs c_double points (got {points.dtype}). Converting.")
            points = points.astype(c_double)
        np.array([num_blocks, n, min_tag, max_tag], dtype=c_size_t).tofile(fh)
    else:
        fh.write(f"{num_blocks} {n} {min_tag} {max_tag}\n".encode())

    for j in range(num_blocks):
        dim, tag = node_dim_tags[j]

        node_tags = np.where(reverse_index_map == j)[0]
        num_points_this = node_tags.size

        if binary:
            np.array([dim, tag, is_parametric], dtype=c_int).tofile(fh)
            np.array([num_points_this], dtype=c_size_t).tofile(fh)
            (node_tags + 1).astype(c_size_t).tofile(fh)
            points[node_tags].tofile(fh)
        else:
            fh.write(f"{dim} {tag} {is_parametric} {num_points_this}\n".encode())
            (node_tags + 1).astype(c_size_t).tofile(fh, "\n", "%d")
            fh.write(b"\n")
            np.savetxt(fh, points[node_tags], delimiter=" ", fmt="%" + float_fmt)

    if binary:
        fh.write(b"\n")

    fh.write(b"$EndNodes\n")


def _write_elements(fh, cells, tag_data, binary: bool) -> None:
    """write the $Elements block

    $Elements
      numEntityBlocks(size_t)
      numElements(size_t) minElementTag(size_t) maxElementTag(size_t)
      entityDim(int) entityTag(int) elementType(int; see below) numElementsInBlock(size_t)
        elementTag(size_t) nodeTag(size_t) ...
        ...
      ...
    $EndElements
    """
    fh.write(b"$Elements\n")

    total_num_cells = sum(len(c) for c in cells)
    num_blocks = len(cells)
    min_element_tag = 1
    max_element_tag = total_num_cells
    if binary:
        np.array(
            [num_blocks, total_num_cells, min_element_tag, max_element_tag],
            dtype=c_size_t,
        ).tofile(fh)

        tag0 = 1
        for ci, cell_block in enumerate(cells):
            node_idcs = _meshio_to_gmsh_order(cell_block.type, cell_block.data)
            if node_idcs.dtype != c_size_t:
                # Binary Gmsh needs c_size_t. Converting."
                node_idcs = node_idcs.astype(c_size_t)

            # entityDim(int) entityTag(int) elementType(int)
            # numElementsBlock(size_t)

            # The entity tag should be equal within a CellBlock
            if "gmsh:geometrical" in tag_data:
                entity_tag = tag_data["gmsh:geometrical"][ci][0]
            else:
                entity_tag = 0

            cell_type = _meshio_to_gmsh_type[cell_block.type]
            np.array([cell_block.dim, entity_tag, cell_type], dtype=c_int).tofile(fh)
            n = node_idcs.shape[0]
            np.array([n], dtype=c_size_t).tofile(fh)

            if node_idcs.dtype != c_size_t:
                warn(
                    f"Binary Gmsh cells need c_size_t (got {node_idcs.dtype}). "
                    + "Converting."
                )
                node_idcs = node_idcs.astype(c_size_t)

            np.column_stack(
                [
                    np.arange(tag0, tag0 + n, dtype=c_size_t),
                    # increment indices by one to conform with gmsh standard
                    node_idcs + 1,
                ]
            ).tofile(fh)
            tag0 += n

        fh.write(b"\n")
    else:
        fh.write(
            "{} {} {} {}\n".format(
                num_blocks, total_num_cells, min_element_tag, max_element_tag
            ).encode()
        )

        tag0 = 1
        for ci, cell_block in enumerate(cells):
            node_idcs = _meshio_to_gmsh_order(cell_block.type, cell_block.data)

            # entityDim(int) entityTag(int) elementType(int) numElementsBlock(size_t)

            # The entity tag should be equal within a CellBlock
            if "gmsh:geometrical" in tag_data:
                entity_tag = tag_data["gmsh:geometrical"][ci][0]
            else:
                entity_tag = 0

            cell_type = _meshio_to_gmsh_type[cell_block.type]
            n = len(cell_block.data)
            fh.write(f"{cell_block.dim} {entity_tag} {cell_type} {n}\n".encode())
            np.savetxt(
                fh,
                # Gmsh indexes from 1 not 0
                np.column_stack([np.arange(tag0, tag0 + n), node_idcs + 1]),
                "%d",
                " ",
            )
            tag0 += n

    fh.write(b"$EndElements\n")


def _write_periodic(fh, periodic, float_fmt: str, binary: bool) -> None:
    """write the $Periodic block

    specified as

    $Periodic
      numPeriodicLinks(size_t)
      entityDim(int) entityTag(int) entityTagMaster(int)
      numAffine(size_t) value(double) ...
      numCorrespondingNodes(size_t)
        nodeTag(size_t) nodeTagMaster(size_t)
        ...
      ...
    $EndPeriodic

    """

    def tofile(fh, value, dtype, **kwargs):
        ary = np.array(value, dtype=dtype)
        if binary:
            ary.tofile(fh)
        else:
            ary = np.atleast_2d(ary)
            fmt = float_fmt if dtype == c_double else "d"
            fmt = "%" + kwargs.pop("fmt", fmt)
            np.savetxt(fh, ary, fmt=fmt, **kwargs)

    fh.write(b"$Periodic\n")
    tofile(fh, len(periodic), c_size_t)
    for dim, (stag, mtag), affine, slave_master in periodic:
        tofile(fh, [dim, stag, mtag], c_int)
        if affine is None or len(affine) == 0:
            tofile(fh, 0, c_size_t)
        else:
            tofile(fh, len(affine), c_size_t, newline=" ")
            tofile(fh, affine, c_double, fmt=float_fmt)
        slave_master = np.array(slave_master, dtype=c_size_t)
        slave_master = slave_master.reshape(-1, 2)
        slave_master = slave_master + 1  # Add one, Gmsh is 1-based
        tofile(fh, len(slave_master), c_size_t)
        tofile(fh, slave_master, c_size_t)
    if binary:
        fh.write(b"\n")
    fh.write(b"$EndPeriodic\n")



from __future__ import annotations

import shlex

import numpy as np
from numpy.typing import ArrayLike

from .._common import warn
from .._exceptions import ReadError, WriteError

c_int = np.dtype("int32")
c_double = np.dtype("float64")


def _fast_forward_to_end_block(f, block):
    """fast-forward to end of block"""
    # See also https://github.com/nschloe/pygalmesh/issues/34

    for line in f:
        try:
            line = line.decode()
        except UnicodeDecodeError:
            pass
        if line.strip() == f"$End{block}":
            break
    else:
        warn(f"${block} not closed by $End{block}.")


def _fast_forward_over_blank_lines(f):
    is_eof = False
    while True:
        line = f.readline().decode()
        if not line:
            is_eof = True
            break
        elif len(line.strip()) > 0:
            break
    return line, is_eof


def _read_physical_names(f, field_data):
    line = f.readline().decode()
    num_phys_names = int(line)
    for _ in range(num_phys_names):
        line = shlex.split(f.readline().decode())
        key = line[2]
        value = np.array(line[1::-1], dtype=int)
        field_data[key] = value
    _fast_forward_to_end_block(f, "PhysicalNames")


def _read_data(f, tag, data_dict, data_size, is_ascii):
    # Read string tags
    num_string_tags = int(f.readline().decode())
    string_tags = [
        f.readline().decode().strip().replace('"', "") for _ in range(num_string_tags)
    ]
    # The real tags typically only contain one value, the time.
    # Discard it.
    num_real_tags = int(f.readline().decode())
    for _ in range(num_real_tags):
        f.readline()
    num_integer_tags = int(f.readline().decode())
    integer_tags = [int(f.readline().decode()) for _ in range(num_integer_tags)]
    num_components = integer_tags[1]
    num_items = integer_tags[2]
    if is_ascii:
        data = np.fromfile(f, count=num_items * (1 + num_components), sep=" ").reshape(
            (num_items, 1 + num_components)
        )
        # The first entry is the node number
        data = data[:, 1:]
    else:
        # binary
        dtype = [("index", c_int), ("values", c_double, (num_components,))]
        data = np.fromfile(f, count=num_items, dtype=dtype)
        if not (data["index"] == range(1, num_items + 1)).all():
            raise ReadError()
        data = np.ascontiguousarray(data["values"])

    _fast_forward_to_end_block(f, tag)

    # The gmsh format cannot distinguish between data of shape (n,) and (n, 1).
    # If shape[1] == 1, cut it off.
    if data.shape[1] == 1:
        data = data[:, 0]

    data_dict[string_tags[0]] = data


# Translate meshio types to gmsh codes
# http://gmsh.info//doc/texinfo/gmsh.html#MSH-file-format-version-2
_gmsh_to_meshio_type = {
    1: "line",
    2: "triangle",
    3: "quad",
    4: "tetra",
    5: "hexahedron",
    6: "wedge",
    7: "pyramid",
    8: "line3",
    9: "triangle6",
    10: "quad9",
    11: "tetra10",
    12: "hexahedron27",
    13: "wedge18",
    14: "pyramid14",
    15: "vertex",
    16: "quad8",
    17: "hexahedron20",
    18: "wedge15",
    19: "pyramid13",
    21: "triangle10",
    23: "triangle15",
    25: "triangle21",
    26: "line4",
    27: "line5",
    28: "line6",
    29: "tetra20",
    30: "tetra35",
    31: "tetra56",
    36: "quad16",
    37: "quad25",
    38: "quad36",
    42: "triangle28",
    43: "triangle36",
    44: "triangle45",
    45: "triangle55",
    46: "triangle66",
    47: "quad49",
    48: "quad64",
    49: "quad81",
    50: "quad100",
    51: "quad121",
    62: "line7",
    63: "line8",
    64: "line9",
    65: "line10",
    66: "line11",
    71: "tetra84",
    72: "tetra120",
    73: "tetra165",
    74: "tetra220",
    75: "tetra286",
    90: "wedge40",
    91: "wedge75",
    92: "hexahedron64",
    93: "hexahedron125",
    94: "hexahedron216",
    95: "hexahedron343",
    96: "hexahedron512",
    97: "hexahedron729",
    98: "hexahedron1000",
    106: "wedge126",
    107: "wedge196",
    108: "wedge288",
    109: "wedge405",
    110: "wedge550",
}
_meshio_to_gmsh_type = {v: k for k, v in _gmsh_to_meshio_type.items()}


def _gmsh_to_meshio_order(cell_type: str, idx: ArrayLike) -> np.ndarray:
    # Gmsh cells are mostly ordered like VTK, with a few exceptions:
    meshio_ordering = {
        # fmt: off
        "tetra10": [0, 1, 2, 3, 4, 5, 6, 7, 9, 8],
        "hexahedron20": [
            0, 1, 2, 3, 4, 5, 6, 7, 8, 11, 13,
            9, 16, 18, 19, 17, 10, 12, 14, 15,
        ],  # https://vtk.org/doc/release/4.2/html/classvtkQuadraticHexahedron.html and https://gmsh.info/doc/texinfo/gmsh.html#Node-ordering
        "hexahedron27": [
            0, 1, 2, 3, 4, 5, 6, 7, 8, 11, 13,
            9, 16, 18, 19, 17, 10, 12, 14, 15,
            22, 23, 21, 24, 20, 25, 26,
        ],
        "wedge15": [
            0, 1, 2, 3, 4, 5, 6, 9, 7, 12, 14, 13, 8, 10, 11
        ],  # http://davis.lbl.gov/Manuals/VTK-4.5/classvtkQuadraticWedge.html and https://gmsh.info/doc/texinfo/gmsh.html#Node-ordering
        "pyramid13": [0, 1, 2, 3, 4, 5, 8, 10, 6, 7, 9, 11, 12],
        # fmt: on
    }
    idx = np.asarray(idx)
    if cell_type not in meshio_ordering:
        return idx
    return idx[:, meshio_ordering[cell_type]]


def _meshio_to_gmsh_order(cell_type: str, idx: ArrayLike) -> np.ndarray:
    # Gmsh cells are mostly ordered like VTK, with a few exceptions:
    gmsh_ordering = {
        # fmt: off
        "tetra10": [0, 1, 2, 3, 4, 5, 6, 7, 9, 8],
        "hexahedron20": [
            0, 1, 2, 3, 4, 5, 6, 7, 8, 11, 16,
            9, 17, 10, 18, 19, 12, 15, 13, 14,
        ],
        "hexahedron27": [
            0, 1, 2, 3, 4, 5, 6, 7, 8, 11, 16,
            9, 17, 10, 18, 19, 12, 15, 13, 14,
            24, 22, 20, 21, 23, 25, 26,
        ],
        "wedge15": [
            0, 1, 2, 3, 4, 5, 6, 8, 12, 7, 13, 14, 9, 11, 10,
        ],
        "pyramid13": [0, 1, 2, 3, 4, 5, 8, 9, 6, 10, 7, 11, 12],
        # fmt: on
    }
    idx = np.asarray(idx)
    if cell_type not in gmsh_ordering:
        return idx
    return idx[:, gmsh_ordering[cell_type]]


def _write_physical_names(fh, field_data):
    # Write physical names
    entries = []
    for phys_name in field_data:
        try:
            phys_num, phys_dim = field_data[phys_name]
            phys_num, phys_dim = int(phys_num), int(phys_dim)
            entries.append((phys_dim, phys_num, phys_name))
        except (ValueError, TypeError):
            warn("Field data contains entry that cannot be processed.")
    entries.sort()
    if entries:
        fh.write(b"$PhysicalNames\n")
        fh.write(f"{len(entries)}\n".encode())
        for entry in entries:
            fh.write('{} {} "{}"\n'.format(*entry).encode())
        fh.write(b"$EndPhysicalNames\n")


def _write_data(fh, tag, name, data, binary):
    fh.write(f"${tag}\n".encode())
    # <http://gmsh.info/doc/texinfo/gmsh.html>:
    # > Number of string tags.
    # > gives the number of string tags that follow. By default the first
    # > string-tag is interpreted as the name of the post-processing view and
    # > the second as the name of the interpolation scheme. The interpolation
    # > scheme is provided in the $InterpolationScheme section (see below).
    fh.write(f"{1}\n".encode())
    fh.write(f'"{name}"\n'.encode())
    fh.write(f"{1}\n".encode())
    fh.write(f"{0.0}\n".encode())
    # three integer tags:
    fh.write(f"{3}\n".encode())
    # time step
    fh.write(f"{0}\n".encode())
    # number of components
    num_components = data.shape[1] if len(data.shape) > 1 else 1
    if num_components not in [1, 3, 9]:
        raise WriteError("Gmsh only permits 1, 3, or 9 components per data field.")

    # Cut off the last dimension in case it's 1. This avoids problems with
    # writing the data.
    if len(data.shape) > 1 and data.shape[1] == 1:
        data = data[:, 0]

    fh.write(f"{num_components}\n".encode())
    # num data items
    fh.write(f"{data.shape[0]}\n".encode())
    # actually write the data
    if binary:
        if num_components == 1:
            dtype = [("index", c_int), ("data", c_double)]
        else:
            dtype = [("index", c_int), ("data", c_double, num_components)]
        tmp = np.empty(len(data), dtype=dtype)
        tmp["index"] = 1 + np.arange(len(data))
        tmp["data"] = data
        tmp.tofile(fh)
        fh.write(b"\n")
    else:
        fmt = " ".join(["{}"] + ["{!r}"] * num_components) + "\n"
        # TODO unify
        if num_components == 1:
            for k, x in enumerate(data):
                fh.write(fmt.format(k + 1, x).encode())
        else:
            for k, x in enumerate(data):
                fh.write(fmt.format(k + 1, *x).encode())

    fh.write(f"$End{tag}\n".encode())



import numpy as np

from .._common import warn
from .._helpers import read, reader_map


def add_args(parser):
    parser.add_argument("infile", type=str, help="mesh file to be read from")
    parser.add_argument(
        "--input-format",
        "-i",
        type=str,
        choices=sorted(list(reader_map.keys())),
        help="input file format",
        default=None,
    )


def info(args):
    # read mesh data
    mesh = read(args.infile, file_format=args.input_format)
    print(mesh)

    # check if the cell arrays are consistent with the points
    is_consistent = True
    for cells in mesh.cells:
        if np.any(cells.data > mesh.points.shape[0]):
            warn("Inconsistent mesh. Cells refer to nonexistent points.")
            is_consistent = False
            break

    # check if there are redundant points
    if is_consistent:
        point_is_used = np.zeros(mesh.points.shape[0], dtype=bool)
        for cells in mesh.cells:
            point_is_used[cells.data] = True
        if np.any(~point_is_used):
            warn("Some points are not part of any cell.")

    return 0



# Key is the enum value of each keyword
# values follow the design of GmfKwdFmt array of
# https://github.com/LoicMarechal/libMeshb/blob/master/sources/libmeshb7.c
# For each keyword we assign whether there is a counter associated with it (second column) and
# its format as a string of numbers
#
# i:integer, f:float, d:dimension
medit_codes = {
    0: ("GmfReserved", "", ""),
    1: ("GmfMeshVersionFormatted", "", "i"),
    2: ("GmfReserved", "", ""),
    3: ("GmfDimension", "", "i"),
    4: ("GmfVertices", "i", "dri"),
    5: ("GmfEdges", "i", "iii"),
    6: ("GmfTriangles", "i", "iiii"),
    7: ("GmfQuadrilaterals", "i", "iiiii"),
    8: ("GmfTetrahedra", "i", "iiiii"),
    9: ("GmfPrisms", "i", "iiiiiii"),
    10: ("GmfHexahedra", "i", "iiiiiiiii"),
    11: ("GmfReserved", "", ""),
    12: ("GmfReserved", "", ""),
    13: ("GmfCorners", "i", "i"),
    14: ("GmfRidges", "i", "i"),
    15: ("GmfRequiredVertices", "i", "i"),
    16: ("GmfRequiredEdges", "i", "i"),
    17: ("GmfRequiredTriangles", "i", "i"),
    18: ("GmfRequiredQuadrilaterals", "i", "i"),
    19: ("GmfTangentAtEdgeVertices", "i", "iii"),
    20: ("GmfNormalAtVertices", "i", "ii"),
    21: ("GmfNormalAtTriangleVertices", "i", "iii"),
    22: ("GmfNormalAtQuadrilateralVertices", "i", "iiii"),
    23: ("GmfAngleOfCornerBound", "", "r"),
    24: ("GmfTrianglesP2", "i", "iiiiiii"),
    25: ("GmfEdgesP2", "i", "iiii"),
    26: ("GmfSolAtPyramids", "i", "sr"),
    27: ("GmfQuadrilateralsQ2", "i", "iiiiiiiiii"),
    28: ("GmfISolAtPyramids", "i", "iiiii"),
    29: ("GmfSubDomainFromGeom", "i", "iii"),
    30: ("GmfTetrahedraP2", "i", "iiiiiiiiiii"),
    31: ("GmfFault_NearTri", "i", "i"),
    32: ("GmfFault_Inter", "i", "i"),
    33: ("GmfHexahedraQ2", "i", "iiiiiiiiiiiiiiiiiiiiiiiiiiii"),
    34: ("GmfExtraVerticesAtEdges", "i", "in"),
    35: ("GmfExtraVerticesAtTriangles", "i", "in"),
    36: ("GmfExtraVerticesAtQuadrilaterals", "i", "in"),
    37: ("GmfExtraVerticesAtTetrahedra", "i", "in"),
    38: ("GmfExtraVerticesAtPrisms", "i", "in"),
    39: ("GmfExtraVerticesAtHexahedra", "i", "in"),
    40: ("GmfVerticesOnGeometricVertices", "i", "ii"),
    41: ("GmfVerticesOnGeometricEdges", "i", "iirr"),
    42: ("GmfVerticesOnGeometricTriangles", "i", "iirrr"),
    43: ("GmfVerticesOnGeometricQuadrilaterals", "i", "iirrr"),
    44: ("GmfEdgesOnGeometricEdges", "i", "ii"),
    45: ("GmfFault_FreeEdge", "i", "i"),
    46: ("GmfPolyhedra", "i", "iiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiii"),
    47: ("GmfPolygons", "", "iiiiiiiii"),
    48: ("GmfFault_Overlap", "i", "i"),
    49: ("GmfPyramids", "i", "iiiiii"),
    50: ("GmfBoundingBox", "", "drdr"),
    51: ("GmfReserved", "", ""),
    52: ("GmfPrivateTable", "i", "i"),
    53: ("GmfFault_BadShape", "i", "i"),
    54: ("GmfEnd", "", ""),
    55: ("GmfTrianglesOnGeometricTriangles", "i", "ii"),
    56: ("GmfTrianglesOnGeometricQuadrilaterals", "i", "ii"),
    57: ("GmfQuadrilateralsOnGeometricTriangles", "i", "ii"),
    58: ("GmfQuadrilateralsOnGeometricQuadrilaterals", "i", "ii"),
    59: ("GmfTangents", "i", "dr"),
    60: ("GmfNormals", "i", "dr"),
    61: ("GmfTangentAtVertices", "i", "ii"),
    62: ("GmfSolAtVertices", "i", "sr"),
    63: ("GmfSolAtEdges", "i", "sr"),
    64: ("GmfSolAtTriangles", "i", "sr"),
    65: ("GmfSolAtQuadrilaterals", "i", "sr"),
    66: ("GmfSolAtTetrahedra", "i", "sr"),
    67: ("GmfSolAtPrisms", "i", "sr"),
    68: ("GmfSolAtHexahedra", "i", "sr"),
    69: ("GmfDSolAtVertices", "i", "sr"),
    70: ("GmfISolAtVertices", "i", "i"),
    71: ("GmfISolAtEdges", "i", "ii"),
    72: ("GmfISolAtTriangles", "i", "iii"),
    73: ("GmfISolAtQuadrilaterals", "i", "iiii"),
    74: ("GmfISolAtTetrahedra", "i", "iiii"),
    75: ("GmfISolAtPrisms", "i", "iiiiii"),
    76: ("GmfISolAtHexahedra", "i", "iiiiiiii"),
    77: ("GmfIterations", "", "i"),
    78: ("GmfTime", "", "r"),
    79: ("GmfFault_SmallTri", "i", "i"),
    80: ("GmfCoarseHexahedra", "i", "i"),
    81: ("GmfComments", "i", "c"),
    82: ("GmfPeriodicVertices", "i", "ii"),
    83: ("GmfPeriodicEdges", "i", "ii"),
    84: ("GmfPeriodicTriangles", "i", "ii"),
    85: ("GmfPeriodicQuadrilaterals", "i", "ii"),
    86: ("GmfPrismsP2", "i", "iiiiiiiiiiiiiiiiiii"),
    87: ("GmfPyramidsP2", "i", "iiiiiiiiiiiiiii"),
    88: ("GmfQuadrilateralsQ3", "i", "iiiiiiiiiiiiiiiii"),
    89: ("GmfQuadrilateralsQ4", "i", "iiiiiiiiiiiiiiiiiiiiiiiiii"),
    90: ("GmfTrianglesP3", "i", "iiiiiiiiiii"),
    91: ("GmfTrianglesP4", "i", "iiiiiiiiiiiiiiii"),
    92: ("GmfEdgesP3", "i", "iiiii"),
    93: ("GmfEdgesP4", "i", "iiiiii"),
    94: ("GmfIRefGroups", "i", "ciii"),
    95: ("GmfDRefGroups", "i", "iii"),
    96: ("GmfTetrahedraP3", "i", "iiiiiiiiiiiiiiiiiiiii"),
    97: ("GmfTetrahedraP4", "i", "iiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiii"),
    98: (
        "GmfHexahedraQ3",
        "i",
        "iiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiii",
    ),
    99: (
        "GmfHexahedraQ4",
        "i",
        "iiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiii",
    ),
    100: ("GmfPyramidsP3", "i", "iiiiiiiiiiiiiiiiiiiiiiiiiiiiiii"),
    101: (
        "GmfPyramidsP4",
        "i",
        "iiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiii",
    ),
    102: ("GmfPrismsP3", "i", "iiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiii"),
    103: (
        "GmfPrismsP4",
        "i",
        "iiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiiii",
    ),
    104: ("GmfHOSolAtEdgesP1", "i", "hr"),
    105: ("GmfHOSolAtEdgesP2", "i", "hr"),
    106: ("GmfHOSolAtEdgesP3", "i", "hr"),
    107: ("GmfHOSolAtTrianglesP1", "i", "hr"),
    108: ("GmfHOSolAtTrianglesP2", "i", "hr"),
    109: ("GmfHOSolAtTrianglesP3", "i", "hr"),
    110: ("GmfHOSolAtQuadrilateralsQ1", "i", "hr"),
    111: ("GmfHOSolAtQuadrilateralsQ2", "i", "hr"),
    112: ("GmfHOSolAtQuadrilateralsQ3", "i", "hr"),
    113: ("GmfHOSolAtTetrahedraP1", "i", "hr"),
    114: ("GmfHOSolAtTetrahedraP2", "i", "hr"),
    115: ("GmfHOSolAtTetrahedraP3", "i", "hr"),
    116: ("GmfHOSolAtPyramidsP1", "i", "hr"),
    117: ("GmfHOSolAtPyramidsP2", "i", "hr"),
    118: ("GmfHOSolAtPyramidsP3", "i", "hr"),
    119: ("GmfHOSolAtPrismsP1", "i", "hr"),
    120: ("GmfHOSolAtPrismsP2", "i", "hr"),
    121: ("GmfHOSolAtPrismsP3", "i", "hr"),
    122: ("GmfHOSolAtHexahedraQ1", "i", "hr"),
    123: ("GmfHOSolAtHexahedraQ2", "i", "hr"),
    124: ("GmfHOSolAtHexahedraQ3", "i", "hr"),
    125: ("GmfBezierBasis", "", "i"),
    126: ("GmfByteFlow", "i", "i"),
    127: ("GmfEdgesP2Ordering", "i", "i"),
    128: ("GmfEdgesP3Ordering", "i", "i"),
    129: ("GmfTrianglesP2Ordering", "i", "iii"),
    130: ("GmfTrianglesP3Ordering", "i", "iii"),
    131: ("GmfQuadrilateralsQ2Ordering", "i", "ii"),
    132: ("GmfQuadrilateralsQ3Ordering", "i", "ii"),
    133: ("GmfTetrahedraP2Ordering", "i", "iiii"),
    134: ("GmfTetrahedraP3Ordering", "i", "iiii"),
    135: ("GmfPyramidsP2Ordering", "i", "iii"),
    136: ("GmfPyramidsP3Ordering", "i", "iii"),
    137: ("GmfPrismsP2Ordering", "i", "iiii"),
    138: ("GmfPrismsP3Ordering", "i", "iiii"),
    139: ("GmfHexahedraQ2Ordering", "i", "iii"),
    140: ("GmfHexahedraQ3Ordering", "i", "iii"),
    141: ("GmfEdgesP1Ordering", "i", "i"),
    142: ("GmfEdgesP4Ordering", "i", "i"),
    143: ("GmfTrianglesP1Ordering", "i", "iii"),
    144: ("GmfTrianglesP4Ordering", "i", "iii"),
    145: ("GmfQuadrilateralsQ1Ordering", "i", "ii"),
    146: ("GmfQuadrilateralsQ4Ordering", "i", "ii"),
    147: ("GmfTetrahedraP1Ordering", "i", "iiii"),
    148: ("GmfTetrahedraP4Ordering", "i", "iiii"),
    149: ("GmfPyramidsP1Ordering", "i", "iii"),
    150: ("GmfPyramidsP4Ordering", "i", "iii"),
    151: ("GmfPrismsP1Ordering", "i", "iiii"),
    152: ("GmfPrismsP4Ordering", "i", "iiii"),
    153: ("GmfHexahedraQ1Ordering", "i", "iii"),
    154: ("GmfHexahedraQ4Ordering", "i", "iii"),
    155: ("GmfFloatingPointPrecision", "", "i"),
    156: ("GmfHOSolAtEdgesP4", "i", "hr"),
    157: ("GmfHOSolAtTrianglesP4", "i", "hr"),
    158: ("GmfHOSolAtQuadrilateralsQ4", "i", "hr"),
    159: ("GmfHOSolAtTetrahedraP4", "i", "hr"),
    160: ("GmfHOSolAtPyramidsP4", "i", "hr"),
    161: ("GmfHOSolAtPrismsP4", "i", "hr"),
    162: ("GmfHOSolAtHexahedraQ4", "i", "hr"),
    163: ("GmfHOSolAtEdgesP1NodesPositions", "i", "rr"),
    164: ("GmfHOSolAtEdgesP2NodesPositions", "i", "rr"),
    165: ("GmfHOSolAtEdgesP3NodesPositions", "i", "rr"),
    166: ("GmfHOSolAtEdgesP4NodesPositions", "i", "rr"),
    167: ("GmfHOSolAtTrianglesP1NodesPositions", "i", "rrr"),
    168: ("GmfHOSolAtTrianglesP2NodesPositions", "i", "rrr"),
    169: ("GmfHOSolAtTrianglesP3NodesPositions", "i", "rrr"),
    170: ("GmfHOSolAtTrianglesP4NodesPositions", "i", "rrr"),
    171: ("GmfHOSolAtQuadrilateralsQ1NodesPositions", "i", "rr"),
    172: ("GmfHOSolAtQuadrilateralsQ2NodesPositions", "i", "rr"),
    173: ("GmfHOSolAtQuadrilateralsQ3NodesPositions", "i", "rr"),
    174: ("GmfHOSolAtQuadrilateralsQ4NodesPositions", "i", "rr"),
    175: ("GmfHOSolAtTetrahedraP1NodesPositions", "i", "rrrr"),
    176: ("GmfHOSolAtTetrahedraP2NodesPositions", "i", "rrrr"),
    177: ("GmfHOSolAtTetrahedraP3NodesPositions", "i", "rrrr"),
    178: ("GmfHOSolAtTetrahedraP4NodesPositions", "i", "rrrr"),
    179: ("GmfHOSolAtPyramidsP1NodesPositions", "i", "rrr"),
    180: ("GmfHOSolAtPyramidsP2NodesPositions", "i", "rrr"),
    181: ("GmfHOSolAtPyramidsP3NodesPositions", "i", "rrr"),
    182: ("GmfHOSolAtPyramidsP4NodesPositions", "i", "rrr"),
    183: ("GmfHOSolAtPrismsP1NodesPositions", "i", "rrrr"),
    184: ("GmfHOSolAtPrismsP2NodesPositions", "i", "rrrr"),
    185: ("GmfHOSolAtPrismsP3NodesPositions", "i", "rrrr"),
    186: ("GmfHOSolAtPrismsP4NodesPositions", "i", "rrrr"),
    187: ("GmfHOSolAtHexahedraQ1NodesPositions", "i", "rrr"),
    188: ("GmfHOSolAtHexahedraQ2NodesPositions", "i", "rrr"),
    189: ("GmfHOSolAtHexahedraQ3NodesPositions", "i", "rrr"),
    190: ("GmfHOSolAtHexahedraQ4NodesPositions", "i", "rrr"),
    191: ("GmfEdgesReferenceElement", "", "rr"),
    192: ("GmfTriangleReferenceElement", "", "rrrrrr"),
    193: ("GmfQuadrilateralReferenceElement", "", "rrrrrrrr"),
    194: ("GmfTetrahedronReferenceElement", "", "rrrrrrrrrrrr"),
    195: ("GmfPyramidReferenceElement", "", "rrrrrrrrrrrrrrr"),
    196: ("GmfPrismReferenceElement", "", "rrrrrrrrrrrrrrrrrr"),
    197: ("GmfHexahedronReferenceElement", "", "rrrrrrrrrrrrrrrrrrrrrrrr"),
    198: ("GmfBoundaryLayers", "i", "iii"),
}



"""
CGNS <https://cgns.github.io/>

TODO link to specification?
"""
import numpy as np

from .._exceptions import ReadError
from .._helpers import register_format
from .._mesh import Mesh


def read(filename):
    import h5py

    f = h5py.File(filename, "r")

    if "Base" not in f:
        raise ReadError('Expected "Base" in file. Malformed CGNS?')

    if "Zone1" not in f["Base"]:
        raise ReadError('Expected "Zone1" in "Base". Malformed CGNS?')

    x = f["Base"]["Zone1"]["GridCoordinates"]["CoordinateX"][" data"]
    y = f["Base"]["Zone1"]["GridCoordinates"]["CoordinateY"][" data"]
    z = f["Base"]["Zone1"]["GridCoordinates"]["CoordinateZ"][" data"]
    points = np.column_stack([x, y, z])

    # f["Base"]["Zone1"]["GridElements"]["ElementRange"][" data"])
    idx_min, idx_max = f["Base"]["Zone1"]["GridElements"]["ElementRange"][" data"]
    data = f["Base"]["Zone1"]["GridElements"]["ElementConnectivity"][" data"]
    cells = np.array(data).reshape(idx_max, -1) - 1

    # TODO how to distinguish cell types?
    if cells.shape[1] != 4:
        raise ReadError("Can only read tetrahedra.")
    cells = [("tetra", cells)]

    return Mesh(points, cells)


def write(filename, mesh, compression="gzip", compression_opts=4):
    import h5py

    f = h5py.File(filename, "w")

    base = f.create_group("Base")

    # TODO something is missing here

    zone1 = base.create_group("Zone1")
    coords = zone1.create_group("GridCoordinates")

    # write points
    coord_x = coords.create_group("CoordinateX")
    coord_x.create_dataset(
        " data",
        data=mesh.points[:, 0],
        compression=compression,
        compression_opts=compression_opts,
    )
    coord_y = coords.create_group("CoordinateY")
    coord_y.create_dataset(
        " data",
        data=mesh.points[:, 1],
        compression=compression,
        compression_opts=compression_opts,
    )
    coord_z = coords.create_group("CoordinateZ")
    coord_z.create_dataset(
        " data",
        data=mesh.points[:, 2],
        compression=compression,
        compression_opts=compression_opts,
    )

    # write cells
    # TODO write cells other than tetra
    elems = zone1.create_group("GridElements")
    rnge = elems.create_group("ElementRange")
    for cell_block in mesh.cells:
        if cell_block.type == "tetra":
            rnge.create_dataset(
                " data",
                data=[1, cell_block.data.shape[0]],
                compression=compression,
                compression_opts=compression_opts,
            )
    conn = elems.create_group("ElementConnectivity")
    for cell_block in mesh.cells:
        if cell_block.type == "tetra":
            conn.create_dataset(
                " data",
                data=cell_block.data.reshape(-1) + 1,
                compression=compression,
                compression_opts=compression_opts,
            )


register_format("cgns", [".cgns"], read, {"cgns": write})



from .common import _gmsh_to_meshio_type as gmsh_to_meshio_type
from .common import _meshio_to_gmsh_type as meshio_to_gmsh_type
from .main import read, write

__all__ = ["read", "write", "gmsh_to_meshio_type", "meshio_to_gmsh_type"]



from __future__ import annotations

import sys
from pathlib import Path

import numpy as np
from numpy.typing import ArrayLike

from ._common import error, num_nodes_per_cell
from ._exceptions import ReadError, WriteError
from ._files import is_buffer
from ._mesh import CellBlock, Mesh

extension_to_filetypes = {}
reader_map = {}
_writer_map = {}


def register_format(
    format_name: str, extensions: list[str], reader, writer_map
) -> None:
    for ext in extensions:
        if ext not in extension_to_filetypes:
            extension_to_filetypes[ext] = []
        extension_to_filetypes[ext].append(format_name)

    if reader is not None:
        reader_map[format_name] = reader

    _writer_map.update(writer_map)


def deregister_format(format_name: str):
    for value in extension_to_filetypes.values():
        if format_name in value:
            value.remove(format_name)

    if format_name in reader_map:
        reader_map.pop(format_name)

    if format_name in _writer_map:
        _writer_map.pop(format_name)


def _filetypes_from_path(path: Path) -> list[str]:
    ext = ""
    out = []
    for suffix in reversed(path.suffixes):
        ext = (suffix + ext).lower()
        try:
            out += extension_to_filetypes[ext]
        except KeyError:
            pass

    if not out:
        raise ReadError(f"Could not deduce file format from path '{path}'.")
    return out


def read(filename, file_format: str | None = None):
    """Reads an unstructured mesh with added data.

    :param filenames: The files/PathLikes to read from.
    :type filenames: str

    :returns mesh{2,3}d: The mesh data.
    """
    if is_buffer(filename, "r"):
        return _read_buffer(filename, file_format)

    return _read_file(Path(filename), file_format)


def _read_buffer(filename, file_format: str | None):
    if file_format is None:
        raise ReadError("File format must be given if buffer is used")
    if file_format == "tetgen":
        raise ReadError(
            "tetgen format is spread across multiple files "
            "and so cannot be read from a buffer"
        )
    if file_format not in reader_map:
        raise ReadError(f"Unknown file format '{file_format}'")

    return reader_map[file_format](filename)


def _read_file(path: Path, file_format: str | None):
    if not path.exists():
        raise ReadError(f"File {path} not found.")

    if file_format:
        possible_file_formats = [file_format]
    else:
        # deduce possible file formats from extension
        possible_file_formats = _filetypes_from_path(path)

    for file_format in possible_file_formats:
        if file_format not in reader_map:
            raise ReadError(f"Unknown file format '{file_format}' of '{path}'.")

        try:
            return reader_map[file_format](str(path))
        except ReadError as e:
            print(e)

    if len(possible_file_formats) == 1:
        msg = f"Couldn't read file {path} as {possible_file_formats[0]}"
    else:
        lst = ", ".join(possible_file_formats)
        msg = f"Couldn't read file {path} as either of {lst}"

    error(msg)
    sys.exit(1)


def write_points_cells(
    filename,
    points: ArrayLike,
    cells: dict[str, ArrayLike] | list[tuple[str, ArrayLike] | CellBlock],
    point_data: dict[str, ArrayLike] | None = None,
    cell_data: dict[str, list[ArrayLike]] | None = None,
    field_data=None,
    point_sets: dict[str, ArrayLike] | None = None,
    cell_sets: dict[str, list[ArrayLike]] | None = None,
    file_format: str | None = None,
    **kwargs,
):
    points = np.asarray(points)
    mesh = Mesh(
        points,
        cells,
        point_data=point_data,
        cell_data=cell_data,
        field_data=field_data,
        point_sets=point_sets,
        cell_sets=cell_sets,
    )
    mesh.write(filename, file_format=file_format, **kwargs)


def write(filename, mesh: Mesh, file_format: str | None = None, **kwargs):
    """Writes mesh together with data to a file.

    :params filename: File to write to.
    :type filename: str

    :params point_data: Named additional point data to write to the file.
    :type point_data: dict
    """
    if is_buffer(filename, "r"):
        if file_format is None:
            raise WriteError("File format must be supplied if `filename` is a buffer")
        if file_format == "tetgen":
            raise WriteError(
                "tetgen format is spread across multiple files, and so cannot be written to a buffer"
            )
    else:
        path = Path(filename)
        if not file_format:
            # deduce possible file formats from extension
            file_formats = _filetypes_from_path(path)
            # just take the first one
            file_format = file_formats[0]

    try:
        writer = _writer_map[file_format]
    except KeyError:
        formats = sorted(list(_writer_map.keys()))
        raise WriteError(f"Unknown format '{file_format}'. Pick one of {formats}")

    # check cells for sanity
    for cell_block in mesh.cells:
        key = cell_block.type
        value = cell_block.data
        if key in num_nodes_per_cell:
            if value.shape[1] != num_nodes_per_cell[key]:
                raise WriteError(
                    f"Unexpected cells array shape {value.shape} for {key} cells. "
                    + f"Expected shape [:, {num_nodes_per_cell[key]}]."
                )
        else:
            # we allow custom keys <https://github.com/nschloe/meshio/issues/501> and
            # cannot check those
            pass

    # Write
    return writer(filename, mesh, **kwargs)




import numpy as np

# from PySide6 import QtCore, QtGui, QtWidgets
# import PySide6.QtCharts as QtCharts

# import logging
# logger = logging.getLogger(__name__)


class ContourAnalysis():
    """Summary

    Attributes:
        canvas (TYPE): Description
        figure (TYPE): Description
        raw_coordinates (list): contour points as tuples
        toolbar (TYPE): Description
    """
    def __init__(self, canvas=False):
        pass

    @staticmethod
    def getCurvature(spline_data):
        """Curvature and radius of curvature of a parametric curve

        der1 is dx/dt and dy/dt at each point
        der2 is d2x/dt2 and d2y/dt2 at each point

        Returns:
            float: Tuple of numpy arrays carrying gradient of the curve,
                   the curvature, radiusses of curvature circles and
                   curvature circle centers for each point of the curve
        """

        coo = spline_data[0]
        der1 = spline_data[3]
        der2 = spline_data[4]

        xd = der1[0]
        yd = der1[1]
        x2d = der2[0]
        y2d = der2[1]
        n = xd**2 + yd**2
        d = xd*y2d - yd*x2d

        # gradient dy/dx = dy/du / dx/du
        gradient = der1[1] / der1[0]

        # radius of curvature
        R = n**(3./2.) / abs(d)

        # curvature
        C = d / n**(3./2.)

        # coordinates of curvature-circle center points
        xc = coo[0] - R * yd / np.sqrt(n)
        yc = coo[1] + R * xd / np.sqrt(n)

        return [gradient, C, R, xc, yc]

    @staticmethod
    def getLeRadius(spline_data, curvature_data):
        """Identify leading edge radius, i.e. smallest radius of
        parametric curve

        Returns:
            FLOAT: leading edge radius, its center and related contour
            point and id
        """

        radius = curvature_data[2]
        rc = np.min(radius)
        # numpy where returns a tuple
        # we take the first element, which is type array
        le_id = np.where(radius == rc)[0]
        # convert the numpy array to a list and take the first element
        le_id = le_id.tolist()[0]
        # leading edge curvature circle center
        xc = curvature_data[3][le_id]
        yc = curvature_data[4][le_id]
        xr, yr = spline_data[0]
        xle = xr[le_id]
        yle = yr[le_id]

        return rc, xc, yc, xle, yle, le_id

    def analyze(self):
        """get specific curve properties"""

        if not self.mainwindow.airfoil.spline_data:
            self.mainwindow.slots.messageBox('Please do splining first')
            return

        spline_data = self.mainwindow.airfoil.spline_data
        curvature_data = ContourAnalysis.getCurvature(spline_data)

        # add new attributes to airfoil instance
        self.mainwindow.airfoil.curvature_data = curvature_data

        self.drawContour()

    # def drawContour(self, quantity='gradient'):
    #     """quantity is one of 'gradient', 'curvature', 'radius' """

    #     spline_data = self.mainwindow.airfoil.spline_data
    #     curvature_data = self.mainwindow.airfoil.curvature_data

    #     selector = {'gradient': 0, 'curvature': 1, 'radius': 2}

    #     points = [QtCore.QPointF(x, y) for x, y in zip(spline_data[0][0],
    #         curvature_data[selector[quantity]])]

    #     self.lineSeries = QtCharts.QLineSeries()
    #     self.lineSeries.append(points)
    #     self.chart.removeAllSeries()
    #     self.chart.addSeries(self.lineSeries)
    #     self.chart.setAxisX(self.axisX, self.lineSeries)
    #     self.chart.setAxisY(self.axisY, self.lineSeries)




import copy

import numpy as np

from foilmesh.utils import Utils

import logging
logger = logging.getLogger(__name__)


class Elliptic:
    def __init__(self, ulines):

        self.ulines = ulines
        
        self.nx = np.array(self.ulines).shape[1]
        self.ny = np.array(self.ulines).shape[0]

        # initialize coordinate array
        self.x = np.empty((self.nx, self.ny))
        self.y = np.empty_like(self.x)

        
        # map coordinates from ulines to i,j index
        self.mapUlines()
 
    def mapUlines(self):
        # FIXME try using np.stack
        for j, uline in enumerate(self.ulines):
            for i, u in enumerate(uline):
                self.x[i, j] = u[0]
                self.y[i, j] = u[1]
    
    def mapToUlines(self):
        self.new_ulines = list()
        for j, uline in enumerate(self.ulines):
            new_uline = list()
            for i, u in enumerate(uline):
                new_uline.append((self.xn[i, j], self.yn[i, j]))
            self.new_ulines.append(new_uline)

    @staticmethod
    def curveNormals(x, y, closed=False):
        istart = 0
        iend = 0
        n = list()

        for i, _ in enumerate(x):

            if closed:
                if i == len(x) - 1:
                    iend = -i - 1
            else:
                if i == 0:
                    istart = 1
                if i == len(x) - 1:
                    iend = -1

            a = np.array([x[i + 1 + iend] - x[i - 1 + istart],
                          y[i + 1 + iend] - y[i - 1 + istart]])
            e = Utils.unit_vector(a)
            n.append([e[1], -e[0]])
            istart = 0
            iend = 0
        return np.array(n)

    def smooth(self, iterations=10, tolerance=1e-3, bnd_type=None, verbose=False):

        self.mapUlines()

        self.xn = copy.deepcopy(self.x)
        self.yn = copy.deepcopy(self.y)

        # calculate normals at boundaries
        # used for Neumann boundary conditions
        normals_left = self.curveNormals(self.xn[0, :], self.yn[0, :])
        normals_right = self.curveNormals(self.xn[-1, :], self.yn[-1, :])
        normals_top = self.curveNormals(self.xn[:, -1], self.yn[:, -1])
        normals_bottom = self.curveNormals(self.xn[:, 0], self.yn[:, 0])

        for iteration in range(iterations):

            # loop internal nodes, index a[0, 0] refers to the first internal node
            for i in range(1, self.nx - 1):
                for j in range(1, self.ny - 1):

                    # g22
                    alpha = 1./4. * ( (self.x[i, j+1] - self.x[i, j-1])**2 \
                                  + (self.y[i, j+1] - self.y[i, j-1])**2 )
                    # g11
                    gamma = 1./4. * ( (self.x[i+1, j] - self.x[i-1, j])**2 \
                                  + (self.y[i+1, j] - self.y[i-1, j])**2 )
                    # g12
                    beta = 1./16. * ( ( self.x[i+1, j] - self.x[i-1, j] ) \
                                    * ( self.x[i, j+1] - self.x[i, j-1] ) \
                                    + ( self.y[i+1, j] - self.y[i-1, j] ) \
                                    * ( self.y[i, j+1] - self.y[i, j-1] ) )
                    # calculate new x-coordinate
                    self.xn[i,j] = -0.5 / (alpha + gamma + 1.e-9) \
                        * (2. * beta * ( self.x[i+1, j+1] - self.x[i-1, j+1] \
                        - self.x[i+1, j-1] + self.x[i-1, j-1] ) \
                        - alpha * ( self.x[i+1, j] + self.x[i-1, j] ) \
                        - gamma * ( self.x[i, j+1] + self.x[i, j-1] ) )
                    # calculate new y-coordinate
                    self.yn[i,j] = -0.5 / (alpha + gamma + 1.e-9) \
                        * (2. * beta * ( self.y[i+1, j+1] - self.y[i-1, j+1] \
                        - self.y[i+1, j-1] + self.y[i-1, j-1] ) \
                        - alpha * ( self.y[i+1, j] + self.y[i-1, j] ) \
                        - gamma * ( self.y[i, j+1] + self.y[i, j-1] ) )

                    # Neumann boundary conditions (normal to boundary here)
                    # project vector a (boundary node to internal node)
                    # onto vector b (normal vector at boundary) and move 
                    # internal node to this position
                    if bnd_type == 'Neumann':
                        if j == 1:
                            a = [self.xn[i,1] - self.xn[i,0],
                                 self.yn[i,1] - self.yn[i,0]]
                            b = normals_bottom[i]
                            projected = np.dot(a, b) / np.dot(b, b) * b
                            self.xn[i,1] = self.xn[i,0] + projected[0]
                            self.yn[i,1] = self.yn[i,0] + projected[1]
                        elif i == self.nx - 1:
                            pass
                        elif j == 1:
                            pass
                        elif j == self.ny - 1:
                            pass

            tol = np.max(np.abs(self.xn - self.x)) + np.max(np.abs(self.yn - self.y))

            if verbose:
                logger.info(f'Iteration={iteration+1:3d}, residual={tol:.3e}')

            if tol < tolerance:
                break

            # update coordinates for next iteration
            self.x = copy.deepcopy(self.xn)
            self.y = copy.deepcopy(self.yn)
                    
        # map coordinates back to uline data structure
        self.mapToUlines()

        return self.new_ulines


from . import (
    _cli,
    abaqus,
    ansys,
    avsucd,
    cgns,
    dolfin,
    exodus,
    flac3d,
    gmsh,
    h5m,
    hmf,
    mdpa,
    med,
    medit,
    nastran,
    netgen,
    neuroglancer,
    obj,
    off,
    permas,
    ply,
    stl,
    su2,
    svg,
    tecplot,
    tetgen,
    ugrid,
    vtk,
    vtu,
    wkt,
    xdmf,
)
from .__about__ import __version__
from ._exceptions import ReadError, WriteError
from ._helpers import (
    deregister_format,
    extension_to_filetypes,
    read,
    register_format,
    write,
    write_points_cells,
)
from ._mesh import CellBlock, Mesh

__all__ = [
    "abaqus",
    "ansys",
    "avsucd",
    "cgns",
    "dolfin",
    "exodus",
    "flac3d",
    "gmsh",
    "h5m",
    "hmf",
    "mdpa",
    "med",
    "medit",
    "nastran",
    "netgen",
    "neuroglancer",
    "obj",
    "off",
    "permas",
    "ply",
    "stl",
    "su2",
    "svg",
    "tecplot",
    "tetgen",
    "ugrid",
    "vtk",
    "vtu",
    "wkt",
    "xdmf",
    "_cli",
    "read",
    "write",
    "register_format",
    "deregister_format",
    "write_points_cells",
    "extension_to_filetypes",
    "Mesh",
    "CellBlock",
    "ReadError",
    "WriteError",
    "topological_dimension",
    "__version__",
]



"""
I/O for FLAC3D format.
"""
from __future__ import annotations

import re
import struct
import time

import numpy as np

from ..__about__ import __version__ as version
from .._common import warn
from .._exceptions import ReadError
from .._files import open_file
from .._helpers import register_format
from .._mesh import Mesh

meshio_only = {
    "zone": {
        "tetra": "tetra",
        "tetra10": "tetra",
        "pyramid": "pyramid",
        "pyramid13": "pyramid",
        "wedge": "wedge",
        "wedge12": "wedge",
        "wedge15": "wedge",
        "wedge18": "wedge",
        "hexahedron": "hexahedron",
        "hexahedron20": "hexahedron",
        "hexahedron24": "hexahedron",
        "hexahedron27": "hexahedron",
    },
    "face": {
        "triangle": "triangle",
        "triangle6": "triangle",
        "triangle7": "triangle",
        "quad": "quad",
        "quad8": "quad",
        "quad9": "quad",
    },
}


numnodes_to_meshio_type = {
    "zone": {4: "tetra", 5: "pyramid", 6: "wedge", 8: "hexahedron"},
    "face": {3: "triangle", 4: "quad"},
}


meshio_to_flac3d_type = {
    "triangle": "T3",
    "quad": "Q4",
    "tetra": "T4",
    "pyramid": "P5",
    "wedge": "W6",
    "hexahedron": "B8",
}


flac3d_to_meshio_order = {
    "triangle": [0, 1, 2],
    "quad": [0, 1, 2, 3],
    "tetra": [0, 1, 2, 3],
    "pyramid": [0, 1, 4, 2, 3],
    "wedge": [0, 1, 3, 2, 4, 5],
    "hexahedron": [0, 1, 4, 2, 3, 6, 7, 5],
}


meshio_to_flac3d_order = {
    "triangle": [0, 1, 2],
    "quad": [0, 1, 2, 3],
    "tetra": [0, 1, 2, 3],
    "pyramid": [0, 1, 3, 4, 2],
    "wedge": [0, 1, 3, 2, 4, 5],
    "hexahedron": [0, 1, 3, 4, 2, 7, 5, 6],
}


meshio_to_flac3d_order_2 = {
    "tetra": [0, 2, 1, 3],
    "pyramid": [0, 3, 1, 4, 2],
    "wedge": [0, 2, 3, 1, 5, 4],
    "hexahedron": [0, 3, 1, 4, 2, 5, 7, 6],
}


flag_to_numdim = {
    "zone": 3,
    "face": 2,
}


def _merge(a: dict, b: dict) -> dict:
    return {**a, **b}


def read(filename):
    """Read FLAC3D f3grid grid file."""
    # Read a small block of the file to assess its type
    # See <https://code.activestate.com/recipes/173220/>
    with open_file(filename, "rb") as f:
        block = f.read(8)
        binary = b"\x00" in block

    mode = "rb" if binary else "r"
    with open_file(filename, mode) as f:
        out = read_buffer(f, binary)

    return out


def read_buffer(f, binary):
    """Read binary or ASCII file."""
    points = []
    point_ids = {}
    f_cells = []
    z_cells = []
    f_cell_sets = {}
    z_cell_sets = {}
    f_cell_ids = []
    z_cell_ids = []

    pidx = 0
    if binary:
        # Not sure what the first bytes represent, the format might be wrong
        # It does not seem to be useful anyway
        _ = struct.unpack("<2I", f.read(8))

        (num_nodes,) = struct.unpack("<I", f.read(4))
        for pidx in range(num_nodes):
            pid, point = _read_point_binary(f)
            points.append(point)
            point_ids[pid] = pidx

        for flag in ["zone", "face"]:
            if flag == "zone":
                cell_ids = z_cell_ids
                cells = z_cells
                cell_sets = z_cell_sets
            else:
                cell_ids = f_cell_ids
                cells = f_cells
                cell_sets = f_cell_sets

            (num_cells,) = struct.unpack("<I", f.read(4))
            for _ in range(num_cells):
                cell_id, cell = _read_cell_binary(f, point_ids)
                cell_ids.append(cell_id)
                _update_cells(cells, cell, flag)
                # mapper[flag][cid] = [cidx]
                # cidx += 1

            (num_groups,) = struct.unpack("<I", f.read(4))
            for _ in range(num_groups):
                name, slot, data = _read_cell_group_binary(f)
                cell_sets[f"{flag}:{name}:{slot}"] = np.array(data)
    else:
        while True:
            line = f.readline()

            if not line:
                break

            if line.strip() == "":
                continue

            split = line.rstrip().split()

            if split[0] == "G":
                pid, point = _read_point_ascii(split)
                points.append(point)
                point_ids[pid] = pidx
                pidx += 1

            elif split[0] == "Z":
                cell_id, cell = _read_cell_ascii(split, point_ids)
                z_cell_ids.append(cell_id)
                _update_cells(z_cells, cell, "zone")

            elif split[0] == "F":
                cell_id, cell = _read_cell_ascii(split, point_ids)
                f_cell_ids.append(cell_id)
                _update_cells(f_cells, cell, "face")

            elif split[0] == "ZGROUP":
                # ZGROUP "Region 2" SLOT 1
                name, slot, data = _read_cell_group_ascii(f, line)
                # Watch out! data refers to the global cell_ids, so we need to
                # adapt this later.
                z_cell_sets[f"zone:{name}:{slot}"] = np.asarray(data)

            elif split[0] == "FGROUP":
                name, slot, data = _read_cell_group_ascii(f, line)
                # Watch out! data refers to the global cell_ids, so we need to
                # adapt this later.
                f_cell_sets[f"face:{name}:{slot}"] = np.asarray(data)

    cells = f_cells + z_cells

    # enforce int type, empty numpy arrays have type float64
    f_cell_ids = np.asarray(f_cell_ids, dtype=int)
    z_cell_ids = np.asarray(z_cell_ids, dtype=int)
    z_offset = len(f_cell_ids)

    cell_ids = np.concatenate([f_cell_ids, z_cell_ids + z_offset], dtype=np.int64)

    cell_blocks = [
        (key, np.array(indices)[:, flac3d_to_meshio_order[key]])
        for key, indices in cells
    ]

    # sanity check, but not really necessary
    # _, counts = np.unique(z_ cell_ids, return_counts=True)
    # assert np.all(counts == 1), "Zone cell IDs not unique"
    # _, counts = np.unique(f_ cell_ids, return_counts=True)
    # assert np.all(counts == 1), "Zone cell IDs not unique"

    # FLAC3D contains global cell ids. Create an inverse array that maps the
    # global IDs to the running index (0, 1,..., n) that's used in meshio.
    if len(f_cell_ids) > 0:
        f_inv = np.full(np.max(f_cell_ids) + 1, -1)
        f_inv[f_cell_ids] = np.arange(len(f_cell_ids))
        f_cell_sets = {key: f_inv[value] for key, value in f_cell_sets.items()}
    if len(z_cell_ids) > 0:
        z_inv = np.full(np.max(z_cell_ids) + 1, -1)
        z_inv[z_cell_ids] = np.arange(len(z_cell_ids))
        z_cell_sets = {
            key: z_inv[value] + z_offset for key, value in z_cell_sets.items()
        }

    cell_sets = _merge(f_cell_sets, z_cell_sets)

    # cell_sets contains the indices into the global cell list. Since this is
    # split up into blocks, we need to split the cell_sets, too.
    bins = np.cumsum([len(cb[1]) for cb in cell_blocks])
    for key, data in cell_sets.items():
        d = np.digitize(data, bins)
        cell_sets[key] = [data[d == k] for k in range(len(cell_blocks))]

    # assert len(cell_ids) == sum(len(block) for _, block in cell_blocks)

    # also store the cell_ids
    cell_data = {}
    if len(cell_blocks) > 0:
        cell_data = {
            "cell_ids": np.split(
                cell_ids, np.cumsum([len(block) for _, block in cell_blocks][:-1])
            )
        }

    return Mesh(
        points=np.array(points),
        cells=cell_blocks,
        cell_data=cell_data,
        cell_sets=cell_sets,
    )


def _read_point_ascii(buf_or_line):
    """Read point coordinates."""
    pid = int(buf_or_line[1])
    point = [float(l) for l in buf_or_line[2:]]
    return pid, point


def _read_point_binary(buf_or_line):
    """Read point coordinates."""
    pid, x, y, z = struct.unpack("<I3d", buf_or_line.read(28))
    return pid, [x, y, z]


def _read_cell_ascii(buf_or_line, point_ids):
    """Read cell connectivity."""
    cid = int(buf_or_line[2])
    cell = buf_or_line[3:]
    is_b7 = buf_or_line[1] == "B7"
    cell = [point_ids[int(l)] for l in cell]
    if is_b7:
        cell.append(cell[-1])
    return cid, cell


def _read_cell_binary(buf_or_line, point_ids):
    """Read cell connectivity."""
    cid, num_verts = struct.unpack("<2I", buf_or_line.read(8))
    cell = struct.unpack(f"<{num_verts}I", buf_or_line.read(4 * num_verts))
    is_b7 = num_verts == 7
    cell = [point_ids[int(l)] for l in cell]
    if is_b7:
        cell.append(cell[-1])
    return cid, cell


def _read_cell_group_binary(buf_or_line):
    # Group name
    (num_chars,) = struct.unpack("<H", buf_or_line.read(2))
    (name,) = struct.unpack(f"<{num_chars}s", buf_or_line.read(num_chars))
    name = name.decode()

    # Slot name
    (num_chars,) = struct.unpack("<H", buf_or_line.read(2))
    (slot,) = struct.unpack(f"<{num_chars}s", buf_or_line.read(num_chars))
    slot = slot.decode()

    # Zones
    (num_zones,) = struct.unpack("<I", buf_or_line.read(4))
    data = struct.unpack(f"<{num_zones}I", buf_or_line.read(4 * num_zones))
    return name, slot, data


def _read_cell_group_ascii(buf_or_line, line: str):
    # a group line read
    # ```
    # ZGROUP 'group five' SLOT 5
    # ```
    m = re.match(r"^([A-Z]+) *[\'\"](.*?)[\'\"] *([A-Z]+) *(.*?) *$", line)
    if m is None:
        raise ReadError(
            'Expected line of the form\n```\nZGROUP "group name" SLOT 5\n```\n '
            + f"but got \n```\n{line}\n```\n"
        )
    assert m.group(1) in {"ZGROUP", "FGROUP"}
    assert m.group(3) == "SLOT"
    name = m.group(2)
    slot = m.group(4)

    i = buf_or_line.tell()
    line = buf_or_line.readline()
    data = []
    while True:
        line = line.rstrip().split()
        if line and (line[0] not in {"*", "ZGROUP", "FGROUP"}):
            data += [int(l) for l in line]
        else:
            buf_or_line.seek(i)
            break
        i = buf_or_line.tell()
        line = buf_or_line.readline()

    return name, slot, data


def _update_cells(cells, cell, flag):
    """Update cell list."""
    cell_type = numnodes_to_meshio_type[flag][len(cell)]
    if len(cells) > 0 and cell_type == cells[-1][0]:
        cells[-1][1].append(cell)
    else:
        cells.append((cell_type, [cell]))


def split_f_z(mesh):
    # FLAC3D makes a difference between ZONES (3D-cells only) and FACES
    # (2D-cells only). Split cells into zcells and fcells, along with the cell
    # sets etc.
    zcells = []
    fcells = []
    for cell_block in mesh.cells:
        if cell_block.type in meshio_only["zone"]:
            zcells.append(cell_block)
        elif cell_block.type in meshio_only["face"]:
            fcells.append(cell_block)

    zsets = {}
    fsets = {}
    for key, cset in mesh.cell_sets.items():
        zsets[key] = []
        fsets[key] = []
        for cell_block, sblock in zip(mesh.cells, cset):
            zsets[key].append(
                sblock if cell_block.type in meshio_only["zone"] else None
            )
            fsets[key].append(
                sblock if cell_block.type in meshio_only["face"] else None
            )

    # remove the data that is only None
    zsets = {
        key: value
        for key, value in zsets.items()
        if not all(item is None for item in value)
    }
    fsets = {
        key: value
        for key, value in fsets.items()
        if not all(item is None for item in value)
    }

    # Right now, the zsets contain indices into the corresponding cell block.
    # FLAC3D expects _global_ indices. Update.
    cell_block_sizes = [len(cb) for cb in zcells]
    for key, data in zsets.items():
        gid = 0
        for n, block in zip(cell_block_sizes, data):
            block += gid
            gid += n

    # TODO not sure if fcells and zcells share a common global index
    cell_block_sizes = [len(cb) for cb in fcells]
    for key, data in fsets.items():
        gid = 0
        for n, block in zip(cell_block_sizes, data):
            block += gid
            gid += n

    for label, values in zsets.items():
        zsets[label] = np.concatenate(values)
    for label, values in fsets.items():
        fsets[label] = np.concatenate(values)

    # flac3d indices start at 1
    for label, values in zsets.items():
        zsets[label] += 1
    for label, values in fsets.items():
        fsets[label] += 1

    return zcells, fcells, zsets, fsets


def write(filename, mesh: Mesh, float_fmt: str = ".16e", binary: bool = False):
    """Write FLAC3D f3grid grid file."""
    skip = [c.type for c in mesh.cells if c.type not in meshio_only["zone"]]
    if skip:
        warn(f'FLAC3D format only supports 3D cells. Skipping {", ".join(skip)}.')

    # split into face/zone data
    zcells, fcells, zsets, fsets = split_f_z(mesh)

    mode = "wb" if binary else "w"
    with open_file(filename, mode) as f:
        if binary:
            # Don't know what these values represent
            f.write(struct.pack("<2I", 1375135718, 3))
        else:
            f.write(f"* FLAC3D grid produced by meshio v{version}\n")
            f.write(f"* {time.ctime()}\n")

        _write_points(f, mesh.points, binary, float_fmt)
        # Make gid an array such that its value can be persitently altered
        # inside the functions.
        gid = np.array(0)
        #
        cells = _translate_zcells(mesh.points, mesh.cells)
        _write_cells(f, cells, "zone", binary, gid)
        _write_groups(f, mesh.cells, zsets, "zone", binary)
        #
        cells = _translate_fcells(fcells)
        _write_cells(f, cells, "face", binary, gid)
        _write_groups(f, mesh.cells, fsets, "face", binary)


def _write_points(f, points, binary, float_fmt=None):
    """Write points coordinates."""
    if binary:
        f.write(struct.pack("<I", len(points)))
        for i, point in enumerate(points):
            f.write(struct.pack("<I3d", i + 1, *point))
    else:
        f.write("* GRIDPOINTS\n")
        for i, point in enumerate(points):
            fmt = "G\t{:8}\t" + "\t".join(3 * ["{:" + float_fmt + "}"]) + "\n"
            f.write(fmt.format(i + 1, *point))


def _write_cells(f, cells, flag: str, binary: bool, gid):
    """Write cells."""
    if binary:
        f.write(
            struct.pack(
                "<I", sum(len(c[1]) for c in cells if c[0] in meshio_only[flag])
            )
        )
        for _, cdata in cells:
            num_cells, num_verts = cdata.shape
            tmp = np.column_stack(
                (
                    np.arange(1, num_cells + 1) + gid,
                    np.full(num_cells, num_verts),
                    cdata + 1,
                )
            ).astype(int)
            f.write(struct.pack(f"<{(num_verts + 2) * num_cells}I", *tmp.ravel()))
            gid += num_cells
    else:
        entity = "ZONES" if flag == "zone" else "FACES"
        abbrev = entity[0]

        f.write(f"* {entity}\n")
        for ctype, cdata in cells:
            fmt = f"{abbrev} {{}} {{}} " + " ".join(["{}"] * cdata.shape[1]) + "\n"
            for entry in cdata + 1:
                gid += 1
                f.write(fmt.format(meshio_to_flac3d_type[ctype], gid, *entry))


def _write_groups(f, cells, materials, flag, binary) -> None:
    """Write groups."""
    if materials is None:
        if binary:
            f.write(struct.pack("<I", 0))
        return

    # TODO filter materials by zones/faces

    if binary:
        f.write(struct.pack("<I", len(materials)))
        for label, group in materials.items():
            num_chars, num_zones = len(label), len(group)
            fmt = f"<H{num_chars}sH7sI{num_zones}I"
            tmp = [
                num_chars,
                label.encode(),
                7,
                b"Default",  # slot
                num_zones,
                *group,
            ]
            f.write(struct.pack(fmt, *tmp))
    else:
        flg = "ZGROUP" if flag == "zone" else "FGROUP"

        f.write(f"* {flag.upper()} GROUPS\n")
        for label, group in materials.items():
            f.write(f'{flg} "{label}" SLOT 1\n')
            _write_table(f, group)


def _translate_zcells(points, cells):
    """Reorder meshio cells to FLAC3D zones.

    Four first points must form a right-handed coordinate system (outward
    normal vectors). Reorder corner points according to sign of scalar triple
    products.
    """
    # See <https://stackoverflow.com/a/42386330/353337>
    def slicing_summing(a, b, c):
        c0 = b[:, 1] * c[:, 2] - b[:, 2] * c[:, 1]
        c1 = b[:, 2] * c[:, 0] - b[:, 0] * c[:, 2]
        c2 = b[:, 0] * c[:, 1] - b[:, 1] * c[:, 0]
        return a[:, 0] * c0 + a[:, 1] * c1 + a[:, 2] * c2

    zones = []
    for cell_block in cells:
        assert cell_block.type in meshio_only["zone"]

        # Compute scalar triple products
        key = meshio_only["zone"][cell_block.type]
        tmp = points[cell_block.data[:, meshio_to_flac3d_order[key][:4]].T]
        det = slicing_summing(tmp[1] - tmp[0], tmp[2] - tmp[0], tmp[3] - tmp[0])

        # Reorder corner points
        data = np.where(
            (det > 0)[:, None],
            cell_block.data[:, meshio_to_flac3d_order[key]],
            cell_block.data[:, meshio_to_flac3d_order_2[key]],
        )
        zones.append((key, data))

    return zones


def _translate_fcells(cells):
    """Reorder meshio cells to FLAC3D faces."""
    faces = []
    for cell_block in cells:
        ctype, data = cell_block
        assert ctype in meshio_only["face"]

        key = meshio_only["face"][ctype]
        data = cell_block.data[:, meshio_to_flac3d_order[key]]
        faces.append((key, data))

    return faces


def _translate_groups(cells, cell_data, field_data, flag):
    """Convert meshio cell_data to FLAC3D groups."""
    dim = np.concatenate(
        [np.full(len(c.data), 2 if c.type in meshio_only["face"] else 3) for c in cells]
    )
    numdim = flag_to_numdim[flag]
    groups = {
        k: np.nonzero(np.logical_and(cell_data == k, dim == numdim))[0] + 1
        for k in np.unique(cell_data)
    }
    groups = {k: v for k, v in groups.items() if v.size}

    labels = {k: str(k) for k in groups.keys()}
    labels[0] = "None"
    if field_data:
        labels.update(
            {v[0]: k for k, v in field_data.items() if v[1] == flag_to_numdim[flag]}
        )

    return dict(zip(labels.values(), groups.values()))


def _write_table(f, data, ncol: int = 20):
    """Write group data table."""
    nrow = len(data) // ncol
    lines = np.split(data, np.full(nrow, ncol).cumsum())
    for line in lines:
        if len(line):
            f.write(" {}\n".format(" ".join([str(l) for l in line])))


register_format("flac3d", [".f3grid"], read, {"flac3d": write})



"""
I/O for the PLY format, cf.
<https://en.wikipedia.org/wiki/PLY_(file_format)>.
<https://web.archive.org/web/20161221115231/http://www.cs.virginia.edu/~gfx/Courses/2001/Advanced.spring.01/plylib/Ply.txt>.
"""
import collections
import datetime
import re
import sys

import numpy as np

from ..__about__ import __version__
from .._common import warn
from .._exceptions import ReadError, WriteError
from .._files import open_file
from .._helpers import register_format
from .._mesh import CellBlock, Mesh

# Reference dtypes
ply_to_numpy_dtype = {
    # [u]char is often used as [u]int, e.g., from Wikipedia:
    # > The word 'list' indicates that the data is a list of values, the first of which
    # > is the number of entries in the list (represented as a 'uchar' in this case).
    "char": np.int8,
    "uchar": np.uint8,
    "short": np.int16,
    "ushort": np.uint16,
    "int": np.int32,
    "int8": np.int8,
    "int32": np.int32,
    "int64": np.int64,
    "uint": np.uint32,
    "uint8": np.uint8,
    "uint16": np.uint16,
    "uint32": np.uint32,
    "uint64": np.uint64,
    "float": np.float32,
    "float32": np.float32,
    "float64": np.float64,
    "double": np.float64,
}
numpy_to_ply_dtype = {np.dtype(v): k for k, v in ply_to_numpy_dtype.items()}


def cell_type_from_count(count):
    if count == 1:
        return "vertex"
    elif count == 2:
        return "line"
    elif count == 3:
        return "triangle"
    elif count == 4:
        return "quad"

    return "polygon"


def read(filename):
    with open_file(filename, "rb") as f:
        mesh = read_buffer(f)
    return mesh


def _next_line(f):
    # fast forward to the next significant line
    while True:
        line = f.readline().decode().strip()
        if line and line[:7] != "comment":
            break
    return line


def read_buffer(f):
    # assert that the first line reads `ply`
    line = f.readline().decode().strip()
    if line != "ply":
        raise ReadError("Expected ply")

    line = _next_line(f)
    endianness = None
    if line == "format ascii 1.0":
        is_binary = False
    elif line == "format binary_big_endian 1.0":
        is_binary = True
        endianness = ">"
    else:
        if line != "format binary_little_endian 1.0":
            raise ReadError()
        is_binary = True
        endianness = "<"

    # read header
    line = _next_line(f)
    num_verts = 0
    num_cells = 0
    point_data_formats = []
    point_data_names = []
    cell_data_names = []
    cell_data_dtypes = []
    while line != "end_header":
        m_vert = re.match("element vertex (\\d+)", line)
        m_face = re.match("element face (\\d+)", line)
        if line[:8] == "obj_info":
            line = _next_line(f)
        elif m_vert is not None:
            num_verts = int(m_vert.groups()[0])

            # read point data
            line = _next_line(f)
            while line[:8] == "property":
                m = re.match("property (.+) (.+)", line)
                assert m is not None
                point_data_formats.append(m.groups()[0])
                point_data_names.append(m.groups()[1])
                line = _next_line(f)
        elif m_face is not None:
            num_cells = int(m_face.groups()[0])

            if num_cells < 0:
                raise ReadError(f"Expected positive num_cells (got `{num_cells}`.")

            # read property lists
            line = _next_line(f)
            # read cell data
            while line[:8] == "property":
                if line[:13] == "property list":
                    m = re.match("property list (.+) (.+) (.+)", line)
                    assert m is not None
                    cell_data_dtypes.append(tuple(m.groups()[:-1]))
                else:
                    m = re.match("property (.+) (.+)", line)
                    assert m is not None
                    cell_data_dtypes.append(m.groups()[0])
                cell_data_names.append(m.groups()[-1])
                line = _next_line(f)
        else:
            raise ReadError(
                "Expected `element vertex` or `element face` or `obj_info`, "
                f"got `{line}`"
            )

    if is_binary:
        mesh = _read_binary(
            f,
            endianness,
            point_data_names,
            point_data_formats,
            num_verts,
            num_cells,
            cell_data_names,
            cell_data_dtypes,
        )
    else:
        mesh = _read_ascii(
            f,
            point_data_names,
            point_data_formats,
            num_verts,
            num_cells,
            cell_data_names,
            cell_data_dtypes,
        )

    return mesh


def _read_ascii(
    f,
    point_data_names,
    point_data_formats,
    num_verts,
    num_cells,
    cell_data_names,
    cell_dtypes,
):
    assert len(cell_data_names) == len(cell_dtypes)

    # assert that all formats are the same
    # Now read the data
    dtype = np.dtype(
        [
            (name, ply_to_numpy_dtype[fmt])
            for name, fmt in zip(point_data_names, point_data_formats)
        ]
    )
    pd = np.genfromtxt(f, max_rows=num_verts, dtype=dtype)

    # split off coordinate data and additional point data
    verts = []
    k = 0
    if point_data_names[0] == "x":
        verts.append(pd["x"])
        k += 1
    if point_data_names[1] == "y":
        verts.append(pd["y"])
        k += 1
    if point_data_names[2] == "z":
        verts.append(pd["z"])
        k += 1
    verts = np.column_stack(verts)

    point_data = {
        point_data_names[i]: pd[point_data_names[i]]
        for i in range(k, len(point_data_names))
    }
    cell_data = {}

    cell_blocks = []

    for k in range(num_cells):
        line = f.readline().decode().strip()
        data = line.split()

        if k == 0:
            # initialize the cell data arrays
            i = 0
            cell_data = {}
            for name, dtype in zip(cell_data_names, cell_dtypes):
                if name == "vertex_indices":
                    n = int(data[i])
                    i += n + 1
                else:
                    cell_data[name] = collections.defaultdict(list)
                    i += 1

        # go over the line
        i = 0
        n = None
        for name, dtype in zip(cell_data_names, cell_dtypes):
            if name == "vertex_indices":
                idx_dtype, value_dtype = dtype
                n = ply_to_numpy_dtype[idx_dtype](data[i])
                dtype = ply_to_numpy_dtype[value_dtype]
                idx = dtype(data[i + 1 : i + n + 1])
                if len(cell_blocks) == 0 or len(cell_blocks[-1][1][-1]) != n:
                    cell_blocks.append((cell_type_from_count(n), [idx]))
                else:
                    cell_blocks[-1][1].append(idx)
                i += n + 1
            else:
                dtype = ply_to_numpy_dtype[dtype]
                # use n from vertex_indices
                assert n is not None
                cell_data[name][n] += [dtype(data[j]) for j in range(i, i + 1)]
                i += 1

    cell_data = {
        key: [np.array(v) for v in value.values()] for key, value in cell_data.items()
    }

    return Mesh(verts, cell_blocks, point_data=point_data, cell_data=cell_data)


def _read_binary(
    f,
    endianness,
    point_data_names,
    formats,
    num_verts,
    num_cells,
    cell_data_names,
    cell_data_dtypes,
):
    ply_to_numpy_dtype_string = {
        "uchar": "i1",
        "uint": "u4",
        "uint8": "u1",
        "uint16": "u2",
        "uint32": "u4",
        "uint64": "u8",
        "int": "i4",
        "int8": "i1",
        "int32": "i4",
        "int64": "i8",
        "float": "f4",
        "float32": "f4",
        "double": "f8",
    }

    # read point data
    dtype = [
        (name, endianness + ply_to_numpy_dtype_string[fmt])
        for name, fmt in zip(point_data_names, formats)
    ]
    point_data = np.frombuffer(
        f.read(num_verts * np.dtype(dtype).itemsize), dtype=dtype
    )
    verts = np.column_stack([point_data["x"], point_data["y"], point_data["z"]])
    point_data = {
        name: point_data[name]
        for name in point_data_names
        if name not in ["x", "y", "z"]
    }

    # Convert strings to proper numpy dtypes
    dts = [
        (
            endianness + ply_to_numpy_dtype_string[dtype[0]],
            endianness + ply_to_numpy_dtype_string[dtype[1]],
        )
        if isinstance(dtype, tuple)
        else endianness + ply_to_numpy_dtype_string[dtype]
        for dtype in cell_data_dtypes
    ]

    # memoryviews can be sliced and passed around without copying. However, the
    # `bytearray()` call here redundantly copies so that the final output arrays
    # are writeable.
    buffer = memoryview(bytearray(f.read()))
    buffer_position = 0

    cell_data = {}
    for (name, dt) in zip(cell_data_names, dts):
        if isinstance(dt, tuple):
            buffer_increment, cell_data[name] = _read_binary_list(
                buffer[buffer_position:], dt[0], dt[1], num_cells, endianness
            )
        else:
            buffer_increment = np.dtype(dt).itemsize
            cell_data[name] = np.frombuffer(
                buffer[buffer_position : buffer_position + buffer_increment], dtype=dt
            )[0]
        buffer_position += buffer_increment

    cells = cell_data.pop("vertex_indices", [])

    return Mesh(verts, cells, point_data=point_data, cell_data=cell_data)


def _read_binary_list(buffer, count_dtype, data_dtype, num_cells, endianness):
    """Parse a ply ragged list into a :class:`CellBlock` for each change in row
    length. The only way to know how many bytes the list takes up is to parse
    it. Hence this function also returns the number of bytes consumed.
    """
    count_dtype, data_dtype = np.dtype(count_dtype), np.dtype(data_dtype)
    count_itemsize = count_dtype.itemsize
    data_itemsize = data_dtype.itemsize
    byteorder = "little" if endianness == "<" else "big"

    # Firstly, walk the buffer to extract all start and end ids (in bytes) of
    # each row into `byte_starts_ends`. Using `np.fromiter(generator)` is
    # 2-3x faster than list comprehension or manually populating an array with
    # a for loop. This is still very much the bottleneck - might be worth
    # ctype-ing in future?
    def parse_ragged(start, num_cells):
        at = start
        yield at
        for _ in range(num_cells):
            count = int.from_bytes(buffer[at : at + count_itemsize], byteorder)
            at += count * data_itemsize + count_itemsize
            yield at

    # Row `i` is given by `buffer[byte_starts_ends[i]: byte_starts_ends[i+1]]`.
    byte_starts_ends = np.fromiter(parse_ragged(0, num_cells), np.intp, num_cells + 1)

    # Next, find where the row length changes and list the (start, end) row ids
    # of each homogeneous block into `block_bounds`.
    row_lengths = np.diff(byte_starts_ends)
    count_changed_ids = np.nonzero(np.diff(row_lengths))[0] + 1

    block_bounds = []
    start = 0
    for end in count_changed_ids:
        block_bounds.append((start, end))
        start = end
    block_bounds.append((start, len(byte_starts_ends) - 1))

    # Finally, parse each homogeneous block. Constructing an appropriate
    # `block_dtype` to include the initial counts in each row avoids any
    # wasteful copy operations.
    blocks = []
    for (start, end) in block_bounds:
        if start == end:
            # This should only happen if the element was empty to begin with.
            continue
        block_buffer = buffer[byte_starts_ends[start] : byte_starts_ends[end]]
        cells_per_row = (row_lengths[start] - count_itemsize) // data_itemsize
        block_dtype = np.dtype(
            [("count", count_dtype), ("data", data_dtype * cells_per_row)]
        )
        cells = np.frombuffer(block_buffer, dtype=block_dtype)["data"]

        cell_type = cell_type_from_count(cells.shape[1])

        blocks.append(CellBlock(cell_type, cells))

    return byte_starts_ends[-1], blocks


def write(filename, mesh: Mesh, binary: bool = True):  # noqa: C901

    with open_file(filename, "wb") as fh:
        fh.write(b"ply\n")

        if binary:
            fh.write(f"format binary_{sys.byteorder}_endian 1.0\n".encode())
        else:
            fh.write(b"format ascii 1.0\n")

        now = datetime.datetime.now().isoformat()
        fh.write(f"comment Created by meshio v{__version__}, {now}\n".encode())

        # counts
        fh.write(f"element vertex {mesh.points.shape[0]:d}\n".encode())
        #
        dim_names = ["x", "y", "z"]
        # From <https://en.wikipedia.org/wiki/PLY_(file_format)>:
        #
        # > The type can be specified with one of char uchar short ushort int uint float
        # > double, or one of int8 uint8 int16 uint16 int32 uint32 float32 float64.
        #
        # We're adding [u]int64 here.
        type_name_table = {
            np.dtype(np.int8): "int8",
            np.dtype(np.int16): "int16",
            np.dtype(np.int32): "int32",
            np.dtype(np.int64): "int64",
            np.dtype(np.uint8): "uint8",
            np.dtype(np.uint16): "uint16",
            np.dtype(np.uint32): "uint32",
            np.dtype(np.uint64): "uint64",
            np.dtype(np.float32): "float",
            np.dtype(np.float64): "double",
        }
        for k in range(mesh.points.shape[1]):
            type_name = type_name_table[mesh.points.dtype]
            fh.write(f"property {type_name} {dim_names[k]}\n".encode())

        pd = []
        for key, value in mesh.point_data.items():
            if len(value.shape) > 1:
                warn(
                    "PLY writer doesn't support multidimensional point data yet. "
                    f"Skipping {key}."
                )
                continue
            type_name = type_name_table[value.dtype]
            fh.write(f"property {type_name} {key}\n".encode())
            pd.append(value)

        num_cells = 0
        legal_cell_types = ["vertex", "line", "triangle", "quad", "polygon"]
        for cell_block in mesh.cells:
            if cell_block.type in legal_cell_types:
                num_cells += cell_block.data.shape[0]

        if num_cells > 0:
            fh.write(f"element face {num_cells:d}\n".encode())

            # possibly cast down to int32
            # TODO don't alter the mesh data
            has_cast = False
            for k, cell_block in enumerate(mesh.cells):
                if cell_block.data.dtype == np.int64:
                    has_cast = True
                    mesh.cells[k] = CellBlock(
                        cell_block.type, cell_block.data.astype(np.int32)
                    )

            if has_cast:
                warn("PLY doesn't support 64-bit integers. Casting down to 32-bit.")

            # assert that all cell dtypes are equal
            cell_dtype = None
            for cell_block in mesh.cells:
                if cell_dtype is None:
                    cell_dtype = cell_block.data.dtype
                if cell_block.data.dtype != cell_dtype:
                    raise WriteError()

            if cell_dtype is not None:
                ply_type = numpy_to_ply_dtype[cell_dtype]
                fh.write(f"property list uint8 {ply_type} vertex_indices\n".encode())

        # TODO other cell data
        fh.write(b"end_header\n")

        if binary:
            # points and point_data
            out = np.rec.fromarrays([coord for coord in mesh.points.T] + pd)
            fh.write(out.tobytes())

            # cells
            for cell_block in mesh.cells:
                if cell_block.type not in legal_cell_types:
                    warn(
                        f'cell_type "{cell_block.type}" is not supported by PLY format '
                        "- skipping"
                    )
                    continue
                # prepend with count
                d = cell_block.data
                out = np.rec.fromarrays(
                    [np.broadcast_to(np.uint8(d.shape[1]), d.shape[0]), *d.T]
                )
                fh.write(out.tobytes())
        else:
            # vertices
            # np.savetxt(fh, mesh.points, "%r")  # slower
            # out = np.column_stack([mesh.points] + list(mesh.point_data.values()))
            out = np.rec.fromarrays([coord for coord in mesh.points.T] + pd)
            fmt = " ".join(["{}"] * len(out[0]))
            out = "\n".join([fmt.format(*row) for row in out]) + "\n"
            fh.write(out.encode())

            # cells
            for cell_block in mesh.cells:
                if cell_block.type not in legal_cell_types:
                    warn(
                        f'cell_type "{cell_block.type}" is not supported by PLY format '
                        + "- skipping"
                    )
                    continue
                #                if cell_type not in cell_type_to_count.keys():
                #                    continue
                d = cell_block.data
                out = np.column_stack(
                    [np.full(d.shape[0], d.shape[1], dtype=d.dtype), d]
                )
                # savetxt is slower
                # np.savetxt(fh, out, "%d  %d %d %d")
                fmt = " ".join(["{}"] * out.shape[1])
                out = "\n".join([fmt.format(*row) for row in out]) + "\n"
                fh.write(out.encode())


register_format("ply", [".ply"], read, {"ply": write})
